// C Source File
// Created 02/02/2005; 15:18:55


#define NBDIR 128


#include <tigcclib.h>

// Main Function

unsigned char Plane = 0;
unsigned char * Plane0;
unsigned char * Plane1;
unsigned char * Plane00;
unsigned char * Plane10;
unsigned char * Plane01;
unsigned char * Plane11;
unsigned char * Buff;

static inline void m7_GetPlanes()
{
	Plane00 = (unsigned char *)GrayDBufGetHiddenPlane (DARK_PLANE);
	Plane10 = (unsigned char *)GrayDBufGetHiddenPlane (LIGHT_PLANE);
	
	GrayDBufToggle();

	Plane01 = (unsigned char *)GrayDBufGetHiddenPlane (DARK_PLANE);
	Plane11 = (unsigned char *)GrayDBufGetHiddenPlane (LIGHT_PLANE);

	GrayDBufToggle();
}

void m7_GrayEnd()
{
  GrayOff ();
	free(Buff);
}
 
void m7_DBufToggle()
{
	if(Plane)
	{
		Plane0 = Plane00;
		Plane1 = Plane10;
	}
	else
	{
		Plane0 = Plane01;
		Plane1 = Plane11;
	}
	GrayDBufToggle();
	Plane = 1 - Plane;
}

void m7_GrayInit()
{
	// Need to add checks here!
	//#warning "Note:  Add a check here"
	Buff = (unsigned char *)calloc(GRAYDBUFFER_SIZE,1);
	//#warning "Note:  Make sure GrayOn() is successful"
	!GrayOn ();
	GrayDBufInit (Buff);
	m7_GetPlanes();
	m7_DBufToggle();
}

typedef struct LoadedSprite
{
	unsigned short NbScales;			// max number of sprite scaled sizes (including original size)
	unsigned short NbImg;					// number of images for every animation and every direction including mask images and -if grayscaled- second plane images for only one scaled size
	unsigned short NbAnim;				// number of animations (ex : stopped, walking, shooting -> 3 animations)
	unsigned short NbDir;					// number of directions available for drawing the sprite (usually 8)
	unsigned short * NbImgAnim;		// number of images for the motion of each animation (ex : stopped -> 1, walking -> 2 or 3)
	unsigned short * ScaledSizes;	// [Width (8 pix step), Height for minimum size, ... , W, H for maximum size]
	unsigned char * Data;
	unsigned char * ImgData;
}LoadedSprite;

typedef struct DisplayedSprite
{
	unsigned short Xscreen;
	unsigned short Yscreen;
	unsigned short Zscreen;
	unsigned short width;	// Width (8 pix step) of the current image
	unsigned short height; // Height of the current image
	unsigned char * img; // pointer to the displayed image
											// (if grayscaled, it points to the dark plane. Light plane is width*height bytes after)
	unsigned char * mask; // pointer to the displayed image mask
}DisplayedSprite;

typedef struct Sprite
{
	unsigned short id; // ID of the sprite
	short x; // x on map
	short y; // y on map
	unsigned short z; // altitude of the sprite
	unsigned short dir; // current direction [0,NBDIR[
	unsigned char anim; // current animation [0, NbAnim[
	unsigned char imganim; // current image (in animation)
	unsigned short slot; // sprite data slot number
	unsigned short nbScales; // number of scales for this data
	DisplayedSprite dsprite; // displayed sprite info
}Sprite;

LoadedSprite ls;
Sprite sp;

void m7sprites_LoadSprite(FILE * spriteFile)
{
	unsigned short i;
	short cumul=0;
	
	fseek (spriteFile, 2, SEEK_SET);
	fread (&(ls.NbScales), sizeof(short), 1, spriteFile);
	fread (&(ls.NbImg), sizeof(short), 1, spriteFile);
	fread (&(ls.NbAnim), sizeof(short), 1, spriteFile);
	fread (&(ls.NbDir), sizeof(short), 1, spriteFile);


	ls.NbImgAnim = (short *)(spriteFile->fpos);
	fseek(spriteFile, sizeof(short) * ls.NbAnim, SEEK_CUR);

	ls.ScaledSizes = (short *)(spriteFile->fpos);
	fseek(spriteFile, sizeof(short) * 2 * ls.NbScales, SEEK_CUR);

	for (i=0; i<ls.NbAnim; i++)
		cumul += ls.NbImgAnim[i];

	cumul *= 2 * ls.NbDir;
	
	ls.Data = (spriteFile->fpos);
	fseek(spriteFile, cumul, SEEK_CUR);
	
	cumul = 0;
	
	for(i=0; i<ls.NbScales; i++)
	{
		cumul += ls.ScaledSizes[i*2] * ls.ScaledSizes[i*2 + 1];
	}

	cumul *= ls.NbImg;

	ls.ImgData = (spriteFile->fpos);

}


void m7sprites_AddSprite()
{
	sp.id = 0;
	sp.x = 0;
	sp.y = 0;
	sp.z = 0;
	sp.dir = 0;
	sp.anim = 0;
	sp.imganim = 0;
	sp.nbScales = ls.NbScales;
}

void m7sprites_Animate()
{
	sp.imganim++;
	sp.imganim%=ls.NbImgAnim[sp.anim];	
}

void m7sprites_SetDisplayed(unsigned short Xscreen,
														unsigned short Yscreen,
														unsigned short Zscreen,
														unsigned short cameraDir,
														unsigned short dir,
														unsigned short scale)
{
	unsigned short i;
	unsigned short cumul=0;
	unsigned short imgNum;
	unsigned short maskNum;
	
	
	sp.dsprite.Xscreen = Xscreen;
	sp.dsprite.Yscreen = Yscreen;
	sp.dsprite.Zscreen = Zscreen;
	
	sp.dsprite.width = ls.ScaledSizes[scale * 2];
	sp.dsprite.height = ls.ScaledSizes[scale * 2 + 1];
	
	for (i = 0; i<sp.anim ;i++)
		cumul += 2 * ls.NbImgAnim[i] * ls.NbDir;
	cumul += 2 * (sp.imganim + (ls.NbImgAnim[sp.anim] * (((((cameraDir + NBDIR) - dir + (NBDIR/(2*ls.NbDir)))%NBDIR) * ls.NbDir)/NBDIR)));
	imgNum = ls.Data[cumul];
	maskNum = ls.Data[cumul + 1];

	cumul=0;
	for (i=0; i<scale; i++)
		cumul += ls.ScaledSizes[i * 2] * ls.ScaledSizes[i * 2 + 1] * ls.NbImg;

	sp.dsprite.img = ls.ImgData + cumul + sp.dsprite.width * sp.dsprite.height * imgNum;
	sp.dsprite.mask = ls.ImgData + cumul + sp.dsprite.width * sp.dsprite.height * maskNum;
}


void m7sprite_DisplayNonClipped(unsigned char *Plane0, unsigned char *Plane1)
{
	short i, j;
	DisplayedSprite * dsprite = &(sp.dsprite);
	short width = dsprite->width, width2, width3;
	short height = dsprite->height;
	short Xscreen = (dsprite->Xscreen) - (width*4);
	short Yscreen = (dsprite->Yscreen) - height - (dsprite->Zscreen);
	short offset = (Yscreen * 30);
	unsigned short shift = Xscreen % 8, hmshift;
	unsigned char * img0 = dsprite->img;
	unsigned char * img1 = dsprite->img + width * height;
	unsigned char * mask = dsprite->mask;
	BYTE byteMask;
	BYTE byteImg0;
	BYTE byteImg1;
	
	if (Xscreen < 0)
	{
		shift = (8 + shift) % 8;
		offset += (Xscreen-7) / 8;
	}
	else
		offset += (Xscreen / 8);
	 
	if (shift)
		width2 = width + 1;
	else
		width2 = width;

	hmshift = 8 - shift;
	width3 = width2 - 1;

	for (i=0; i<height; i++)
	{
		if (shift)
		{
			byteMask = (*mask >> shift) | (0xFF << (hmshift));
			byteImg0 = (*img0 >> shift);
			byteImg1 = (*img1 >> shift);
		}
		else
		{
			byteMask = *mask;
			byteImg0 = *img0;
			byteImg1 = *img1;
			mask++;
			img0++;
			img1++;
		}

		Plane0[offset] &= byteMask;
		Plane0[offset] |= byteImg0;
		Plane1[offset] &= byteMask;
		Plane1[offset] |= byteImg1;
				
		for (j=1; j<width3; j++)
		{
		
			if (shift)
			{
				byteMask = (*mask << (hmshift)) | (*(mask+1) >> shift);
				byteImg0 = (*img0 << (hmshift)) | (*(img0+1) >> shift);
				byteImg1 = (*img1 << (hmshift)) | (*(img1+1) >> shift);
			}
			else
			{
				byteMask = *mask;
				byteImg0 = *img0;
				byteImg1 = *img1;
			}
			mask++;
			img0++;
			img1++;
			Plane0[offset + j] &= byteMask;
			Plane0[offset + j] |= byteImg0;
			Plane1[offset + j] &= byteMask;
			Plane1[offset + j] |= byteImg1;
		}
		if (j == width3)
		{
			if (shift)
			{
				byteMask = (*mask << (hmshift)) | (0xFF >> shift);
				byteImg0 = (*img0 << (hmshift));
				byteImg1 = (*img1 << (hmshift));
			}
			else
			{
				byteMask = *mask;
				byteImg0 = *img0;
				byteImg1 = *img1;
			}
			mask++;
			img0++;
			img1++;		
			Plane0[offset + j] &= byteMask;
			Plane0[offset + j] |= byteImg0;
			Plane1[offset + j] &= byteMask;
			Plane1[offset + j] |= byteImg1;
		}
		offset+=30;
	}
	
	
}


void _main(void)
{
  ESI argptr;
  int argtype;
  FILE * sprite;
	short c = 0;
	unsigned short scale = 0;
	short x=80, y=50;

	
  InitArgPtr (argptr);
  if ((argtype = GetArgType (argptr)) == END_TAG)
  	return;
	if (argtype != STR_TAG)
		return;
	if ((sprite = fopen (GetStrnArg (argptr), "rb")) == NULL)
		return;
	
	clrscr();
	m7_GrayInit();
	
	m7sprites_LoadSprite(sprite);
	m7sprites_AddSprite();
	
	do
	{
		switch(c)
		{
			case KEY_F1 :
				m7sprites_Animate();				
				break;
			case KEY_F2 :
				sp.anim+=1;
				sp.anim%=ls.NbAnim;
				break;
			case KEY_UP :	
				scale+=sp.nbScales-1;
				scale%=sp.nbScales;
				break;
			case KEY_DOWN :
				scale+=sp.nbScales+1;
				scale%=sp.nbScales;
				break;
			case KEY_LEFT :
				sp.dir+=NBDIR-1;
				sp.dir%=NBDIR;
				break;
			case KEY_RIGHT :
				sp.dir+=NBDIR+1;
				sp.dir%=NBDIR;
				break;
			case '8' :
				y--;
				break;
			case '2' :
				y++;
				break;
			case '4' :
				x--;
				break;
			case '6' :
				x++;
				break;
			default :;
		}
		
		m7sprites_SetDisplayed(x, y, 0, 0, sp.dir, scale);

		memset(Plane0, 0, LCD_SIZE);
		memset(Plane1, 0, LCD_SIZE);
		m7sprite_DisplayNonClipped(Plane0, Plane1);
		m7_DBufToggle();

		memset(Plane0, 0, LCD_SIZE);
		memset(Plane1, 0, LCD_SIZE);
		m7sprite_DisplayNonClipped(Plane0, Plane1);
		m7_DBufToggle();
		
	}while((c = ngetchx()) != KEY_ESC);




	
	
	m7_GrayEnd();
	fclose (sprite);

	
	

}
