// C Source File
// Created 28/04/2005; 21:51:23

#include "game.h"

extern unsigned char SkyChange;

// gameplay settings
#define IMPULSE 3 // max time for an input to consider it as an impulsion
#define INPUT2X 3 // time between two inputs to consider them as a double input
#define SHIELDTIME 7 // shield time during effect
#define SLIDETIME 20 // slide time during effect
#define MAX_BOOST 100
#define ENERGY_BOOST 14
#define ENERGY_UP 3
#define ENERGY_DOWN 3
#define ENERGY_MAX 136 // 2x the size of the energy bar container
#define COLBOXSIZE 5
#define ENERGY_CRITICAL (ENERGY_MAX/3)
#define AI_JUMP_SPEED 6000
#define SLOW_DOWN_MIN 4000
#define WIND_DIRECTION (NBDIR/4)
#define WIND_STRENGTH 700
#define MAX_JUMP_SPEED 6000

// linked player KeyBoard
#define RUP    (recv&0x80)
#define RDOWN  (recv&0x40)
#define RLEFT  (recv&0x20)
#define RRIGHT (recv&0x10)
#define RGO    (recv&0x08)
#define RBREAK (recv&0x04)
#define RBOOST (recv&0x02)
#define RMENU  (recv&0x01)

unsigned char workingMachines[MAX_MACHINES];

void (*d_next)(void);
void (*m_next)(void);
void (*m_hidden)(void);

#define TRANSTIME 24
static unsigned char trans_t = TRANSTIME;
static unsigned char modifOk = 0;

void d_hide(void) {
	memset(Plane0+ 30*OFFSET_SKY_VERT, 0xFF, LCD_SIZE - 30*OFFSET_SKY_VERT);
	memset(Plane1+ 30*OFFSET_SKY_VERT, 0xFF, LCD_SIZE - 30*OFFSET_SKY_VERT);
	if(modifOk) {
		if(modifOk-- == 1) {
			m7_InitMoreDisplay(d_next);
			m7_SetLoopFunc(m_next);
			SkyChange = 2;
		}
	}
}

void m_hiddenWrap(void) {
	if(!modifOk) m_hidden(), modifOk = 2;
}

void m_trans(void) {
	if(!(trans_t--)) {
		m7_InitMoreDisplay(d_hide);
		m7_SetLoopFunc(m_hiddenWrap);
		trans_t = TRANSTIME;
	}
}

void d_trans(void) {
	unsigned char i, *pl0 = Plane0 + 30*OFFSET_SKY_VERT, *pl1 = Plane1 + 30*OFFSET_SKY_VERT, transt = trans_t, linelen = 0;
	for(i=SIZE_Y*ZOOM_SCR_VERT+SIZE_SKY_Y;i--;) {
		if(i%2) {
			linelen = random(4) + (30*(TRANSTIME-transt))/TRANSTIME;
			linelen = (linelen >= 30)?30:linelen;
		}
		memset (pl0, 0xFF, linelen);
		memset (pl1, 0xFF, linelen);
		pl0 +=30;
		pl1 +=30;
	}
}

static inline void transistion(void (*modifFunc)(void), void (*moreDispFunc)(void), void (*hiddenFunc)(void)) {
	m_next = modifFunc;
	d_next = moreDispFunc;
	m_hidden = hiddenFunc;
	m7_InitMoreDisplay(d_trans);
	m7_SetLoopFunc(m_trans);
}

void dummy(void) {}

static inline void KBD_READ(void) {
	if(!linkenabled)
		kbd_Scan();
}

void resetMachines(void) {
	unsigned short i;
	for (i=MAX_MACHINES;i--;)
		workingMachines[i] = 0;
}

void addMachines(void) {
	unsigned short i;

	for (i=nb_machines;i--;) {
		const unsigned short * stattab = stats[sel_machines[i]];
		workingMachines[i] = 1;
		machines[i].id = sprites_id[i];
		machines[i].controlled = controlled_tab[i];

		machines[i].weight = stattab[0]/100;
		machines[i].body = stattab[1];
		machines[i].boost = stattab[2];
		machines[i].grip = stattab[3];
		machines[i].accel = stattab[4] + (120 - balance_tab[i])/12; // max gain 10
		machines[i].maxspeed = stattab[5] + balance_tab[i]/3; // max gain 40
		machines[i].cornering = stattab[6];

		machines[i].energy = ENERGY_MAX;
		machines[i].attacked = 0;
		machines[i].currentwp = 11;
		machines[i].previouswp = 11;
		machines[i].lapcount = 1;
		machines[i].lapreverse = 1;
		machines[i].bestlap = 0xFFFF;
		machines[i].totaltime = 0;
		machines[i].currentlap = 0;
		machines[i].rank = 0;

		
		machines[i].thrust = 0;
		machines[i].altitude = 0;
		machines[i].potential = 0;
		machines[i].risespeed = 0;
		machines[i].fallspeed = 0;
		machines[i].extForce.dir = 0;
		machines[i].extForce.norm = 0;
		machines[i].inertia.dir = 0;
		machines[i].inertia.norm = 0;

		unsigned char j;
		for(j=SLIDE_GRIP;j--;)
			machines[i].dirs[j] = 0;
		machines[i].sliding = 0;
		machines[i].boosting = 0;
		machines[i].turning = 0;
		machines[i].doubleleft = 0;
		machines[i].doubleright = 0;
		machines[i].impulse = 0;
		machines[i].shield = 0;
		machines[i].accelerating = 0;
		
		machines[i].boostperiod = 0;
		machines[i].energyperiod = 0;
		machines[i].jumping = 0;
		
		machines[i].hold = 0;
		machines[i].data = gparam.data;
	}
}

static inline unsigned char bestrank(Machine * m1, Machine * m2) {
	return ((m1->lapcount > m2->lapcount && m2->lapcount) || (!m1->lapcount && m2->lapcount) || (m1->currentwp > m2->currentwp && m1->lapcount == m2->lapcount));
}

static volatile unsigned char RankFinish;

void computeRank(void) {
	unsigned char i, j = 0, k, unfinish[4];
	Machine * machine;
	for(i=nb_machines;i--;) {
		machine = &(machines[i]);
		if(!(rank[i] = machine->rank)) {
			if(!workingMachines[i]) rank[i] = 5; // disqualified (will be 4 after last pass)
			else unfinish[j++] = i, rank[i] = RankFinish++;
		}
	}
	
	for(i=j;i--;) {
		for(k=i;k--;) {
			if (bestrank(&machines[unfinish[i]], &machines[unfinish[k]])) {
				rank[unfinish[i]] ^= rank[unfinish[k]];
				rank[unfinish[k]] ^= rank[unfinish[i]];
				rank[unfinish[i]] ^= rank[unfinish[k]];
			}
		}
	}
	
	for(i=nb_machines;i--;) rank[i]--;
}

#define KUP    (keyboard[0])
#define KDOWN  (keyboard[1])
#define KLEFT  (keyboard[2])
#define KRIGHT (keyboard[3])
#define KGO    (keyboard[4])
#define KBREAK (keyboard[5])
#define KBOOST (keyboard[6])

static volatile unsigned char lap;
static volatile unsigned short tick;
static volatile unsigned short laptick;
static volatile unsigned char boosteventtime;
static volatile unsigned char energy;
static volatile unsigned short speed;
static volatile unsigned char ranka;
static volatile unsigned char playerrev;
static volatile unsigned short playerwp;
static volatile unsigned char mapchangetime;
static volatile unsigned char spritex;
static volatile unsigned char spritey;


static inline void compute_env(void) {
	unsigned char i;
	Vect thrust;
	Machine * machine;
	Sprite * sprite = NULL;
	unsigned char keyboard[7]; //UP DOWN LEFT RIGHT GO BREAK BOOST
	unsigned short tilepos;
	unsigned char tilenum;
	unsigned short waypoints[9];

	// keyboard & gameplay
	for(i=nb_machines;i--;) {
		if(workingMachines[i]) {
			machine = &(machines[i]);
			sprite = AllSpriteList[machine->id];
			unsigned char j, k, kdir, kmain, km1;
			unsigned short gt;
			switch(machine->controlled) {
				case 0: // Player
					KUP    = UP;
					KDOWN  = DOWN;
					KLEFT  = LEFT;
					KRIGHT = RIGHT;
					KGO    = GO;
					KBREAK = BREAK;
					KBOOST = BOOST;
					break;
				default: // CPUs
					for(j=7;j--;) keyboard[j] = 0;
					
					if(machine->inertia.norm < (2+machine->controlled)*1024) KGO = 1;
					
					tilepos = ((sprite->x)>>(TILES_SHIFT + SINCOS + ZOOM)) + ((sprite->y)>>(TILES_SHIFT + SINCOS + ZOOM)) * xm->dimX;
					waypoints[0] = WPM[tilepos];
					tilenum = xm->map[tilepos];
					waypoints[1] = WPM[tilepos+xm->dimX];
					waypoints[2] = WPM[tilepos+xm->dimX+1];
					waypoints[3] = WPM[tilepos+1];
					waypoints[4] = WPM[tilepos-xm->dimX+1];
					waypoints[5] = WPM[tilepos-xm->dimX];
					waypoints[6] = WPM[tilepos-xm->dimX-1];
					waypoints[7] = WPM[tilepos-1];
					waypoints[8] = WPM[tilepos+xm->dimX-1];
					k = 0;
					kmain = 1;
					nokmain:
					kdir = 0;
					km1 = ((sprite->dir/(NBDIR/8))+1);
					gt = waypoints[0]<0x8000?waypoints[0]:0;
					if(!sprite->z)
						for(j=km1;j>0;j--) {
							
							if(kdir && j==km1) continue;
							if(waypoints[0]>10 && waypoints[j]<5) waypoints[j] |= 0x4000;
							if(waypoints[0]<5 && waypoints[j]>10) waypoints[j] = 0x0;
							if(gt < waypoints[j] && waypoints[j] < 0x8000) {
								gt = waypoints[j]; k = j;
							}
							if(!kdir) {
								j = 8; kdir = 1;
							}
						}
					if(!k && kmain) {
						kmain = 0;
						for(j=8;j--;) if(waypoints[j]<0xFFFE) waypoints[j] &= 0x7FFF;
						goto nokmain;
					}
					if(k) { if(sprite->dir != NBDIR/8 * (k-1)) {sprite->dir = NBDIR/8 * (k-1); machine->hold = (7-machine->controlled) * 3; } }
					
					if(machine->boostperiod) machine->boostperiod--;
					else KBOOST = 1, machine->boostperiod = 256/GAME_SPEED * (8-machine->controlled);

					if(machine->energyperiod) machine->energyperiod--;
					else machine->energy = ENERGY_MAX, machine->energyperiod = 256/GAME_SPEED * (8-machine->controlled) * 5;
					
					if(sprite->z) {
						machine->inertia.norm = AI_JUMP_SPEED;
						if(!machine->jumping) {
							machine->potential = AI_JUMP_SPEED;
							machine->risespeed = AI_JUMP_SPEED/machine->weight;
							machine->altitude = 0;
							machine->jumping = 1;
						}
						machine->boosting = 0;
						KBOOST = 0;
					} else machine->jumping = 0;
					
					break;
				case 2: // Ghosts
					if(!(machine->hold--)) {
						recv = *(machine->data++);
						machine->hold = *(machine->data++);
					}
				case 1: // Linked Player
					KUP    = RUP;
					KDOWN  = RDOWN;
					KLEFT  = RLEFT;
					KRIGHT = RRIGHT;
					KGO    = RGO;
					KBREAK = RBREAK;
					KBOOST = RBOOST;
					break;
			}
			
			if(track.mapstyle == 6) {
				machine->extForce.dir = WIND_DIRECTION;
				machine->extForce.norm = WIND_STRENGTH;
			}
			
			if(KLEFT) {
				if((machine->doubleleft) && ((machine->doubleleft) != INPUT2X)) {
					machine->turning = -96;
					(machine->shield) = SHIELDTIME;
				} else if (machine->shield) {
					machine->turning = -96;
					(machine->shield)--;
				} else {
					machine->turning -= machine->cornering;
					if (machine->turning < -64)
						machine->turning = -64;
				}
				if ((machine->impulse) <= IMPULSE && !(machine->shield)) {
					(machine->doubleleft) = INPUT2X;
					(machine->impulse)++;
				}	else
					(machine->doubleleft) = 0;
				(machine->doubleright) = 0;
			}
			else
			if(KRIGHT) {
				if((machine->doubleright) && ((machine->doubleright) != INPUT2X)) {
					machine->turning = 96;
					(machine->shield) = SHIELDTIME;
				}	else if (machine->shield)	{
					machine->turning = 96;
					(machine->shield)--;
				} else {
					machine->turning += machine->cornering;
					if (machine->turning > 64)
						machine->turning = 64;
				}
				if ((machine->impulse) <= IMPULSE && !(machine->shield)) {
					(machine->doubleright) = INPUT2X;
					(machine->impulse)++;
				} else
					(machine->doubleright) = 0;
				(machine->doubleleft) = 0;
			}
			else {
				if (machine->turning >= 32)
					machine->turning -=32;
				else
					machine->turning = 0;

				if (machine->turning <= -32)
					machine->turning +=32;
				else
					machine->turning = 0;

				if(machine->doubleleft)
					(machine->doubleleft)--;
				if(machine->doubleright)
					(machine->doubleright)--;
				(machine->impulse) = 0;
			}
						
			if (machine->turning == -96) {
				if(!machine->controlled) m7sprites_Set1stDisplayed(1, (NBDIR/7 * 3) - 1);
				if (machine->inertia.norm > 0) 
					if ((machine->shield) >= SHIELDTIME - 2) {
						machine->extForce.dir = (sprite->dir + NBDIR/4) % NBDIR;
						machine->extForce.norm = (machine->inertia.norm << 1);
					}
			}
			else
			if (machine->turning <= -64) {
				if(!machine->controlled) m7sprites_Set1stDisplayed(1, (NBDIR/7 * 2) - 1);
				if (machine->inertia.norm > 0) sprite->dir += 3;
			}
			else
			if (machine->turning <= -1) {
				if(!machine->controlled) m7sprites_Set1stDisplayed(1, (NBDIR/7 * 1) - 1);
				if (machine->inertia.norm > 0) sprite->dir += 2;
			}
			else
			if (machine->turning == 0) {
				if(!machine->controlled) m7sprites_Set1stDisplayed(1, 0);
			}
			else
			if (machine->turning == 96) {
				if(!machine->controlled) m7sprites_Set1stDisplayed(1, (NBDIR/7 * 4) - 1);
				if (machine->inertia.norm > 0) 
					if ((machine->shield) >= SHIELDTIME - 2) {
						machine->extForce.dir = (sprite->dir - NBDIR/4 + NBDIR) % NBDIR;
						machine->extForce.norm = (machine->inertia.norm << 1);
					}
			}
			else
			if (machine->turning >= 64) {
				if(!machine->controlled) m7sprites_Set1stDisplayed(1, (NBDIR/7 * 5) - 1);
				if (machine->inertia.norm > 0) sprite->dir -= 3 - NBDIR;
			}
			else
			if (machine->turning >= 1) {
				if(!machine->controlled) m7sprites_Set1stDisplayed(1, (NBDIR/7 * 6) - 1);
				if (machine->inertia.norm > 0) sprite->dir -= 2 - NBDIR;
			}
					
			sprite->dir %= NBDIR;
						
			if(sprite->z == 0) {
				if(KBREAK && machine->turning && !(machine->sliding))
					(machine->sliding) = SLIDETIME + 1;
			}	else {
				if(KDOWN) {
					machine->fallspeed -= 1;// lenghten jump
				}
				else if(KUP) {
					machine->fallspeed += Gravity;// shorten jump
				}
			}
			
			if (machine->sliding)
				(machine->sliding)--;
				
			if(KBREAK) {
				if(!machine->controlled) {
					if (AllSpriteList[3]->anim != 0) {
						AllSpriteList[3]->imganim = 0;
						AllSpriteList[3]->anim = 0;
					}
				}
				if (machine->inertia.norm >	2 * FORCE_DECR)
					machine->inertia.norm -=	2 * FORCE_DECR;
				else
					machine->inertia.norm = 0;
		
				if (machine->thrust > FORCE_DECR)
					machine->thrust -= FORCE_DECR;
				else
					machine->thrust = 0;
			} else {
				if(KGO) {
					if (!(machine->accelerating))
						(machine->sliding) = 0;
					(machine->accelerating) = 1;
					if (machine->boosting) {
						if(!machine->controlled) {
							if (AllSpriteList[3]->anim != 2) {
								AllSpriteList[3]->imganim = 0;
								AllSpriteList[3]->anim = 2;
							}
						}
						if (machine->thrust < (machine->maxspeed + MAX_BOOST))
							machine->thrust += machine->accel << 1;							
					}	else {
						if(!machine->controlled) {
							if (AllSpriteList[3]->anim != 1) {
								AllSpriteList[3]->imganim = 0;
								AllSpriteList[3]->anim = 1;
							}
						}
						if (machine->thrust < machine->maxspeed)
							machine->thrust += machine->accel;
						else
							machine->thrust -= machine->accel << 2;
					}
				} else {
					if(!machine->controlled) {
						if (AllSpriteList[3]->anim != 0) {
							AllSpriteList[3]->imganim = 0;
							AllSpriteList[3]->anim = 0;
						}
					}
					(machine->accelerating) = 0;
					if (machine->thrust > FORCE_DECR)
						machine->thrust -= FORCE_DECR;
					else
						machine->thrust = 0;
				}
			}		
					
			if (machine->boosting)
				(machine->boosting)--;
		
			if (KBOOST && machine->lapcount > 1) {
				if (!(machine->boosting)) {
					if(machine->lapcount) {
						if (machine->energy > ENERGY_BOOST) {
							machine->energy -= ENERGY_BOOST;
							machine->boosting = machine->boost;
						}
						if (!machine->controlled) energy = machine->energy, EnergyChange = 2;
					}
				} else if (!machine->controlled) BOOST = 0;
			}
		}
	}


	// collision detection between machines
	short orientation = 0;
	for(i=nb_machines;i--;) {
		if(workingMachines[i]) {
			machine = &(machines[i]);
			if(machine->controlled != 2) {
				sprite = AllSpriteList[machine->id];
				unsigned char j;
				long xmaci,ymaci,xv,yv;
				xmaci = sprite->x;
				ymaci = sprite->y;
				for(j=i;j--;) {
					if(workingMachines[j]) {
						long xmacy,ymacy;
						xmacy = xmaci-AllSpriteList[machines[j].id]->x;
						ymacy = ymaci-AllSpriteList[machines[j].id]->y;
						xv = xmacy*xmacy;
						yv = ymacy*ymacy;
						xv += yv;
						if(xv < (COLBOXSIZE<<(ZOOM+SINCOS))*(COLBOXSIZE<<(ZOOM+SINCOS))) {
							
							signed short Py;
							signed short proj = 0;
							unsigned short z;
						
							for (z = NBDIR; z--;) {
								Py = (ymacy * CosTable[z] + xmacy * SinTable[z]);
								if (proj < Py) {
									proj = Py;
									orientation = z;
								}
							}
							
							Py = machine->inertia.norm;
							machine = &(machines[j]);
							Py += machine->inertia.norm;
							
							collide_ship (&(machine->inertia), (NBDIR/2 + orientation)%NBDIR, &(machine->extForce), Py);
							if(machine->lapcount && !machine->shield) {
								if (machine->energy > ENERGY_DOWN)
									machine->energy -= ENERGY_DOWN;
								else
									machine->energy = 0;
								if (!machine->controlled) energy = machine->energy, EnergyChange = 2;
							}
							
							machine = &(machines[i]);
							collide_ship (&(machine->inertia), orientation, &(machine->extForce), Py);
							if(machine->lapcount && !machine->shield) {
								if (machine->energy > ENERGY_DOWN)
									machine->energy -= ENERGY_DOWN;
								else
									machine->energy = 0;
								if (!machine->controlled) energy = machine->energy, EnergyChange = 2;
							}
	
						}
					}
				}
			}
		}
	}
	AddedSprites[2] = 3;
	// collision detection with tiles

	for(i=nb_machines;i--;) {
		if(workingMachines[i]) {
			unsigned char effect, param;
			unsigned short tileposmask, maskpos, waypoint, pre, reverse = 0;
			machine = &(machines[i]);
			sprite = AllSpriteList[machine->id];
			
			if(!machine->controlled) {spritex = ((sprite->x)>>(TILES_SHIFT + SINCOS + ZOOM)); spritey = ((sprite->y)>>(TILES_SHIFT + SINCOS + ZOOM));}
			
			if(sprite->z == 0) {
				
				tilepos = ((sprite->x)>>(TILES_SHIFT + SINCOS + ZOOM)) + ((sprite->y)>>(TILES_SHIFT + SINCOS + ZOOM)) * xm->dimX;
				waypoint = WPM[tilepos];
				if(waypoint < 0xFFFE) waypoint &= 0x7FFF;
				else waypoint = machine->currentwp;
				
				// start line detection, lap counter
				if(machine->currentwp != waypoint) {
					pre = machine->previouswp;
					machine->previouswp = machine->currentwp;
					machine->currentwp = waypoint;
					if((pre > waypoint || (pre < 5 && waypoint > 10)) && !(pre > 10 && waypoint < 5)) reverse = 1;
					if((machine->previouswp < 5) && waypoint > 10) { // lapreverse
						machine->lapreverse = 1;
						if(!machine->controlled) playerrev = 1;
						reverse = 1;
					}
					if((machine->previouswp > 10) && waypoint < 5) { // lap ++
						if(machine->lapreverse) { machine->lapreverse = 0; if(!machine->controlled) playerrev = 0; }
						else if(machine->lapcount) {
							machine->lapcount++;
							if(machine->currentlap < machine->bestlap) machine->bestlap = machine->currentlap;
							machine->currentlap = 0;
							if(!machine->controlled) {
								lap = machine->lapcount, LapChange = 2;
								if (lap == 2) {boostevent = 1;boosteventtime = 256/GAME_SPEED;}
							}
						}
					}
					if(!machine->controlled) {
						playerwp = waypoint;
						if(reverse) {
							if(!revevent) revevent = 1;
						} else {
							if(revevent) {
								DrawHUDFull(); HudChange = 2; MapChange = 2; revevent = 0;
								LapChange = 2; EnergyChange = 2; RankChange = 2;
							}
						}
					}
				}
				
				tilenum = xm->map[tilepos];
				tileposmask = ((sprite->x)>>(TILES_SHIFT + SINCOS + ZOOM)) + ((sprite->y)>>(TILES_SHIFT + SINCOS + ZOOM)) * (1 << (10-TILES_SHIFT));
				maskpos = ((((sprite->y)>>(SINCOS + ZOOM)) & (TILES_SIZE - 1)) << TILES_SHIFT) + (((sprite->x)>>(SINCOS + ZOOM)) & (TILES_SIZE - 1));

				if((xm->mask[tileposmask])[maskpos]) {
					if (tilenum >= nbfx)
						effect = 0;
					else
						effect = effects[tilenum];
						
					if (effect >= 0xC0) { // Nothing
						;
					}	else
					if (effect >= 0xA0) { // energy
						if(machine->lapcount) {
							if (machine->energy < ENERGY_MAX - ENERGY_UP)
								machine->energy += ENERGY_UP;
							else
								machine->energy = ENERGY_MAX;
							if (!machine->controlled) { energy = machine->energy, EnergyChange = 2; AddedSprites[2] = 2; }
						}
					} else
					if (effect >= 0x80) { // jumper
						if(machine->controlled >2) {
							switch (tilenum) {
								case 68: machine->inertia.dir = NBDIR/2; break; // up
								case 69: machine->inertia.dir = 3*NBDIR/4; break; // left
								case 70: machine->inertia.dir = 0; break; // down
								case 71: machine->inertia.dir = NBDIR/4; break; // right
								default:;
							}
						}
						machine->potential = machine->inertia.norm;
						if(machine->potential > MAX_JUMP_SPEED) machine->potential = MAX_JUMP_SPEED;
						machine->risespeed = machine->potential/machine->weight;
					}	else
					if (effect >= 0x60) { // ice
						if(machine->controlled < 3) machine->sliding = SLIDETIME + 1;
					} else
					if (effect >= 0x40) { // speed modificator
						param = effect & 0x1F;
						if (param < 3)	{
							machine->inertia.norm -= (machine->inertia.norm >> 2);
							if(param == 2) {
								if(machine->lapcount) {
									if (machine->energy > ENERGY_DOWN)
										machine->energy -= ENERGY_DOWN;
									else
										machine->energy = 0;
									if (!machine->controlled) energy = machine->energy, EnergyChange = 2;
								}
							}
						} else
						if (param == 3) {
							machine->boosting = machine->boost;
						}
					} else
					if (effect >= 0x20) { // collision
						collide (&(machine->inertia), (effect & 0x1F) * (NBDIR/32), &(machine->extForce));
						if(machine->lapcount) {
							if (machine->energy > ENERGY_DOWN)
								machine->energy -= ENERGY_DOWN;
							else
								machine->energy = 0;
							if (!machine->controlled) energy = machine->energy, EnergyChange = 2;
						}
					}
					else { // destruction
						if(machine->lapcount) {
							machine->energy = 0;
							if (!machine->controlled) energy = machine->energy, EnergyChange = 2;
						}
					}
				}
			}
			
		}
	}
	
	unsigned char rankl = 1;

	// forces application on machines
	for(i=nb_machines;i--;) {
		if(workingMachines[i]) {
			machine = &(machines[i]);
			sprite = AllSpriteList[machine->id];

			if(machine->controlled) {
				if ((machine->currentwp > playerwp && machine->lapcount == lap && (machine->lapreverse == playerrev)) ||
						(machine->lapcount > lap) ||
						(!machine->lapcount) ||
						(machine->currentwp <= playerwp && machine->lapcount == lap && playerrev && !machine->lapreverse)) rankl++;
			}

			translate (&(sprite->x), &(sprite->y), &(machine->extForce));
			translate (&(sprite->x), &(sprite->y), &(machine->inertia));
			thrust.norm = machine->thrust;
			thrust.dir = sprite->dir;
			translate (&(sprite->x), &(sprite->y), &(thrust));
			thrustToInertia(&thrust, &(machine->inertia), machine->sliding, machine->dirs);

			if (machine->extForce.norm >= EXTFORCE_DECR)
				machine->extForce.norm -= EXTFORCE_DECR;
			else
				machine->extForce.norm = 0;

			if (machine->inertia.norm >= FORCE_DECR)
				machine->inertia.norm -= FORCE_DECR;
			else
				machine->inertia.norm = 0;

			if (machine->potential >= Gravity*machine->weight) {
				machine->potential -= Gravity*machine->weight;
				machine->altitude += machine->risespeed;
				machine->risespeed -= Gravity;
			}
			else
				machine->potential = 0;
			
			if(machine->altitude >= machine->fallspeed) {
				machine->altitude -= machine->fallspeed;
				machine->fallspeed += Gravity;
			}	else {
				machine->altitude = 0;
				machine->fallspeed = 0;
			}
			
			sprite->z = machine->altitude>>7;
			if(sprite->z > MAX_ALTITUDE) sprite->z = MAX_ALTITUDE;

		}
	}

	if(rankl != ranka) ranka = rankl, RankChange = 2;

}

unsigned char animcritical, shaking;

void m_race(void) {
	unsigned char i;
	unsigned char kbd_status;
	Machine * machine;
	
	for(i=nb_machines;i--;) {
		if(workingMachines[i]) {
			machine = &(machines[i]);
			machine->totaltime++;
			machine->currentlap++;
			if(!machine->controlled) {
				tick = machine->totaltime, laptick = machine->currentlap, speed = machine->inertia.norm >> 3;
			}
			if(machine->energy == 0) {
				workingMachines[i] = 0;
				if(!machine->controlled) end_race_status = 1;
				else AddedSprites[machine->id] = 3;
			}
			if(machine->lapcount > track.nbLaps) {
				machine->lapcount = 0, machine->rank = RankFinish++;// finish
				if(linkenabled) { if(!mparam.nbcpu) machines[1-i].rank = 2; end_race_status = 2;}
				if(machine->controlled == 2) machine->attacked = 1, AddedSprites[machine->id] = 3;
			}
			if(!machine->controlled && !machine->lapcount) end_race_status = 2;
			
		}
	}
	if(boosteventtime) {
		boosteventtime--;
		if(!boosteventtime) {
			DrawHUDFull(); HudChange = 2; MapChange = 2; boostevent = 0;
			LapChange = 2; EnergyChange = 2; RankChange = 2;
		} 
	}
	if(!mapchangetime--) { mapchangetime = 64/GAME_SPEED; MapChange = 2; }
	KBD_READ();
	if(MENU || (linkenabled && RMENU)) end_race_status = 3;
	compute_env();
	if(ghost_record) {
		kbd_status = 0;
		if(UP)kbd_status|=0x80;
		if(DOWN)kbd_status|=0x40;
		if(LEFT)kbd_status|=0x20;
		if(RIGHT)kbd_status|=0x10;
		if(GO)kbd_status|=0x08;
		if(BREAK)kbd_status|=0x04;
		if(BOOST)kbd_status|=0x02;
		if((*(ghost_record_buffer_current-1) != kbd_status) || hold == 0xFF) {
			*ghost_record_buffer_current++ = hold;
			hold = 0;
			*ghost_record_buffer_current++ = kbd_status;
			ghost_record++;
			if(ghost_record == GHOST_BUFFER_SIZE/2) ghost_record = 0;
		} else hold++;
	}
	
	if(machines[1-linkmode].boosting && !AllSpriteList[1]->z) shaking = 2;
	else if(shaking == 1) AllSpriteList[1]->dsprite.Yscreen = SPRITEY, shaking = 0;
	
	if (machines[1-linkmode].energy < ENERGY_CRITICAL) {
		if (AllSpriteList[1]->anim != 1) {
			AllSpriteList[1]->imganim = 0;
			AllSpriteList[1]->anim = 1;
		}
		if (machines[1-linkmode].energy < ENERGY_CRITICAL/3)
			animcritical = 1;
		else
			if (machines[1-linkmode].energy < 2 * ENERGY_CRITICAL/3)
				animcritical = 2;
			else
				animcritical = 3;
	} else {
		if (AllSpriteList[1]->anim != 0) {
			AllSpriteList[1]->imganim = 0;
			AllSpriteList[1]->anim = 0;
		}
	}

	// mode7 camera update
	m7_SetCameraPos2(AllSpriteList[1]->x, AllSpriteList[1]->y);
	m7_SetDir(AllSpriteList[1]->dir);
	m7_Translate(-5);

	AllSpriteList[1]->dsprite.Zscreen = AllSpriteList[1]->z;
	AllSpriteList[3]->dsprite.Zscreen = AllSpriteList[1]->z;

	if(end_race_status)
		m7_LoopEnd();
}

void d_animate_cpu(void) {
	unsigned char i;
	m7sprites_Animate(4);
	if(ghostloaded) {
		if(!machines[1].attacked) AddedSprites[7] = 4-AddedSprites[7];
	} else {
		for(i=nb_machines-1;i--;)
			m7sprites_Animate(7+i);
	}
}

void d_race(void) {
	static unsigned char i, rankc;
	static unsigned char anim = 0, shake = 0;

	if(boostevent == 1) {
		boostevent = 2;
		DrawHUDFrame();
		DrawHUD4P();
		FontSetSys(F_8x10);
		for(i=0;i<4;i++) {
			PortSet (Planes[i], 239, 127);
			DrawStr (LOGOX - 44+(i%2), HUDY+9, "BOOST POWER", A_REVERSE);
		}
	} else if(!boostevent) {
		if(revevent == 1) {
			revevent = 2;
			DrawHUDFrame();
			DrawHUD4P();
			FontSetSys(F_8x10);
			for(i=0;i<4;i++) {
				PortSet (Planes[i], 239, 127);
				DrawStr (LOGOX - 28+(i%2), HUDY+9, "REVERSE", A_REVERSE);
			}
		} else if(!revevent) {
			if(HudChange) {
				DrawHUD2P();
				HudChange--;
			}
			if(MapChange) {
				m7sprite_Display1stPlane(5, Plane0, Plane1);
				PortSet (Plane0, 239, 127);
				spritex /= 3, spritey /= 3;
				DrawPix (HUDXO+24+spritex, HUDY+2+spritey, A_XOR);
				DrawPix (HUDXO+24+spritex, HUDY+2+spritey+1, A_XOR);
				DrawPix (HUDXO+24+spritex+1, HUDY+2+spritey, A_XOR);
				DrawPix (HUDXO+24+spritex+1, HUDY+2+spritey+1, A_XOR);
				MapChange--;
			}
			if(EnergyChange) {
				m7sprite_DisplayNonClipped(6, Plane0, Plane1);
				PortSet (Plane0, 239, 127);
				DrawLine(HUDXO + 121 - 34, HUDY+14-9, HUDXO + 121 - 35 + (energy >> 1), HUDY+14-9, A_NORMAL);
				DrawLine(HUDXO + 121 - 34, HUDY+14-8, HUDXO + 121 - 35 + (energy >> 1), HUDY+14-8, A_NORMAL);
				DrawLine(HUDXO + 121 - 34, HUDY+14-7, HUDXO + 121 - 35 + (energy >> 1), HUDY+14-7, A_NORMAL);
				EnergyChange--;
			}
			if(LapChange) {
				FontSetSys(F_4x6);
				PortSet (Plane0, 239, 127);
				for(i=2; i--;) {
					lapbuff[4] = '0' + lap;
					DrawStr (HUDXO+86, HUDY+17, lapbuff, A_REVERSE);
					PortSet (Plane1, 239, 127);
				}
				LapChange--;
			}
			if(RankChange) {
				FontSetSys(F_8x10);
				PortSet (Plane0, 239, 127);
				for(i=2; i--;) {
					rankc = '0' + ranka;
					DrawChar (HUDXO+6, HUDY+11, rankc, A_REVERSE);
					PortSet (Plane1, 239, 127);
				}
				RankChange--;
			}
			fd_ReverseGrayDrawTime(((HUDXO+55)/4), HUDY+10, tick, Plane0, Plane1);
			fd_ReverseGrayDrawTime(((HUDXO+55)/4), HUDY+17, laptick, Plane0, Plane1);
			fd_ReverseGrayDrawSpeed(((HUDXO+120)/4), HUDY+17, speed, Plane0, Plane1);
		}
		
	}
	
	if (anim >= animcritical) {
		m7sprites_Animate(1);
		m7sprites_Set1stDisplayed(1, 0);
		anim = 0;	
	}
	anim++;

	if(shaking) {
		shaking = 1;
		if (shake)
			AllSpriteList[1]->dsprite.Yscreen = SPRITEY;
		else
			AllSpriteList[1]->dsprite.Yscreen = SPRITEY + 1;
		shake = 1-shake;
	}
	
	m7sprites_Animate(2);
	m7sprites_Set1stDisplayed(2, 0);
	
	m7sprites_Animate(3);
	m7sprites_Set1stDisplayed(3, 0);
	d_animate_cpu();
}

static inline void swap_scaled(void) {
	unsigned short i;
	AddedSprites[0] = 5-AddedSprites[0];
	AddedSprites[1] = 5-AddedSprites[1];
	AddedSprites[3] = 5-AddedSprites[3];
	AddedSprites[4] = 4-AddedSprites[4];
	i = (AddedSprites[4] == 3)?4:1;
	AllSpriteList[5-i]->x = AllSpriteList[i]->x;
	AllSpriteList[5-i]->y = AllSpriteList[i]->y;
	AllSpriteList[5-i]->z = AllSpriteList[i]->z;
	AllSpriteList[5-i]->dir = AllSpriteList[i]->dir;
}

void m_initcam(void) {
	swap_scaled();
	m7_SetCameraPos2(AllSpriteList[1]->x, AllSpriteList[1]->y);
	m7_SetDir(AllSpriteList[1]->dir);
	m7_Translate(-5);
	DrawHUD4P();
}

static unsigned char s321[4] = {'5','\0','!','\0'};

void m_321go(void) {
	if(tick) tick--;
}

void d_321go(void) {
	unsigned char i;
	static unsigned char w=4;
	FontSetSys(F_8x10);
	if(!tick) {
		tick = 256/GAME_SPEED;
		if (w==4) (*s321)--;
	}
	if (*s321 == '0') w = 12, *s321 = 'G', *(s321+1) = 'O';
	if (*s321 != '4') {
		PortSet (Plane0, 239, 127);
		for(i=2; i--;) {
			DrawStr (LOGOX - w+i, HUDY+9, s321, A_REVERSE);
			PortSet (Plane1, 239, 127);
		}
	}
	if (w == 12 && tick <= 128/GAME_SPEED) {
		*s321 = '5';
		*(s321+1) = '\0';
		DrawHUDFull();
		HudChange = 2;
		MapChange = 2;
		lap = 1;
		LapChange = 2;
		EnergyChange = 2;
		revevent = 0;
		boostevent = 0;
		energy = ENERGY_MAX;
		speed = 0;
		RankChange = 2;
		playerrev = 1;
		RankFinish = 1;
		mapchangetime = 10;
		animcritical = 3;
		ghost_record_buffer_current = ghost_record_buffer+1;
		hold = 0xFF;
		ranka = rank[linkenabled && !linkmode];
		m7_InitMoreDisplay(d_race);
		m7_SetLoopFunc(m_race);
		tick = 0;
		w=4;
	}
	m7sprites_Animate(3);
	m7sprites_Set1stDisplayed(3, 0);
	d_animate_cpu();
}

void m_race_intro(void) {
	static short u = 5, t = 192-32*4, s = 64, r = NBDIR/4;

	KBD_READ();
	if(!(MENU || (linkenabled && RMENU))) {
		if (s) {
			m7_Translate(2);
			s--;
			return;
		}
		
		if(r) {
			m7_Translate(t);
			m7_Rotate(4);
			t--;
			m7_Translate(-t);
			r--;
			return;
		}
		
		if(u--)
			return;
	}

	u = 5;
	t = 192-32*4;
	s = 64;
	r = NBDIR/4;
	tick = 0;
	laptick = 0;
	transistion(m_321go, d_321go, m_initcam);
}


void m_explode() {
	static unsigned char anim = 0;
	
	m7sprites_Animate(0);
	m7sprites_Animate(1);
	m7sprites_Animate(2);
	m7sprites_Animate(3);
	m7sprites_Animate(4);
	if (anim & 1)
		m7_Translate(1);
	anim++;
	
	if (anim == 27) {
		anim = 0;
		m7_LoopEnd();
	}
}

#define NBSMOKESPRT 4

void m_crash() {
	static short s = 30, r = NBDIR/4;
	unsigned short i;

	if (s) {
		m7_Translate(1);
		s--;
		return;
	}
	
	for (i = NBSMOKESPRT; i--;) {
		AllSpriteList[i]->z += 2;
		if (AllSpriteList[i]->z >= MAX_ALTITUDE)
			AllSpriteList[i]->z = 0;
	}
	
	if(r) {
		m7_Rotate(2);
		r--;
		return;
	}

	kbd_Scan();

	if(MENU || (linkenabled && RMENU)) {
		s = 30;
		r = NBDIR/4;
		m7_LoopEnd();
	}
}

void gameover(void) {
	unsigned short i;
	unsigned short x, y;
	m7_GetCameraPos(&x, &y);
	
	m7sprites_RemoveAllSprites();
	m7sprites_FreeAllSprite();
	m7sprites_AddSprite(m7sprites_LoadSprite(fzeropk0, "sprites\\shared\\explode"));
	AllSpriteList[0]->x = x;
	AllSpriteList[0]->y = y;
	m7sprites_AddSprite(0);
	AllSpriteList[1]->x = x;
	AllSpriteList[1]->y = y+112;
	m7sprites_AddSprite(0);
	AllSpriteList[2]->x = x+112;
	AllSpriteList[2]->y = y;
	m7sprites_AddSprite(0);
	AllSpriteList[3]->x = x;
	AllSpriteList[3]->y = y-112;
	m7sprites_AddSprite(0);
	AllSpriteList[4]->x = x-112;
	AllSpriteList[4]->y = y;

	DrawHUDFrame();
	DrawHUD4P();
	
	for(i=0;i<4;i++) {
		PortSet (Planes[i], 239, 127);
		FontSetSys(F_8x10);
		DrawStr (LOGOX - 28+(i%2), HUDY+9, "RETIRED", A_REVERSE);
	}
	
	m7_InitMoreDisplay(dummy);
	m7_SetLoopFunc(m_explode);

	m7_LoopStart();

	m7sprites_RemoveAllSprites();
	m7sprites_FreeAllSprite();

	m7sprites_LoadSprite(fzeropk0, "sprites\\shared\\smoke");

	for (i = 0; i < NBSMOKESPRT; i++) {
		m7sprites_AddSprite(0);
		AllSpriteList[i]->x = x + (i & 1) * ((((i >> 1) & 1) << 1) - 1) * 32;
		AllSpriteList[i]->y = y + ((i+1) & 1) * (((((i+1) >> 1) & 1) << 1) - 1) * 32;
		AllSpriteList[i]->z = i * MAX_ALTITUDE / (NBSMOKESPRT - 1);
	}

	if(linkenabled) {
		DrawHUDFrame();
		DrawHUD4P();
		for(i=0;i<4;i++) {
			PortSet (Planes[i], 239, 127);
			FontSetSys(F_6x8);
			DrawStr (LOGOX - 75+(i%2), HUDY+9, "WAITING FOR LINKED PLAYER", A_REVERSE);
		}
	}
	
	m7_InitMoreDisplay(dummy);
	m7_SetLoopFunc(m_crash);

	m7_LoopStart();

	m7sprites_RemoveAllSprites();
	m7sprites_FreeAllSprite();
}

void m_finish(void) {
	static unsigned char nbloop = 60;
	compute_env();
	if(nbloop) nbloop--;
	else nbloop = 60, m7_LoopEnd();
}

void finish(void) {
	swap_scaled();
	unsigned char i;
	Machine * machine;
	for(i=nb_machines;i--;) {
		if(workingMachines[i]) {
			machine = &(machines[i]);
			machine->energy = ENERGY_MAX;
			if(!machine->controlled) { machine->controlled = 3; machine->id = 4; }
		}
	}

	DrawHUDFrame();
	DrawHUD4P();
	
	for(i=0;i<4;i++) {
		PortSet (Planes[i], 239, 127);
		FontSetSys(F_8x10);
		if(linkenabled) {
			if(machines[1-linkmode].rank == 1)
				DrawStr (LOGOX - 28+(i%2), HUDY+9, "YOU WON", A_REVERSE);
			else
				DrawStr (LOGOX - 32+(i%2), HUDY+9, "YOU LOSE", A_REVERSE);
		} else {
			DrawStr (LOGOX - 24+(i%2), HUDY+9, "FINISH", A_REVERSE);
		}
	}
	
	m7_InitMoreDisplay(d_animate_cpu);
	m7_SetLoopFunc(m_finish);
	
	m7_LoopStart();
}
