// C Source File
// Created 17/02/2005; 20:08:24

#include "game.h"

static unsigned char *pl0, *pl1;

static inline void getplanes(void) {
	pl0 = Plane0;
	pl1 = Plane1;					 
}

static inline void clrscrR(void) {
	memset (pl0, 0xFF, LCD_SIZE);
	memset (pl1, 0xFF, LCD_SIZE);
}

static inline void clrscrG(void) {
	memset (pl0, 0x00, LCD_SIZE);
	memset (pl1, 0x00, LCD_SIZE);
}

static inline void clearAllR(void) {
	getplanes();
	clrscrR();
	m7_DBufToggle();
	getplanes();
	clrscrR();
	m7_DBufToggle();
}

static inline void finish_display(unsigned char toggle) {
	PortSet (pl1, 239, 127);
	ScrRectFill (&(SCR_RECT){{LOGOX-60-2, LOGOY-56,LOGOX+60+3, LOGOY+37}}
						 ,&(SCR_RECT){{LOGOX-60-2, LOGOY-56,LOGOX+60+3, LOGOY+37}}, A_XOR);
	if (toggle) m7_DBufToggle();
}

typedef enum {K_UP, K_DOWN, K_LEFT, K_RIGHT, K_ENTER, K_ESC, K_NO_PRESS}MKs;
static unsigned char menu_keys[] = { 0, 0, 0, 0, 0, 0 };

static inline void test_key_mask(unsigned char mask) {
	BEGIN_KEYTEST
		menu_keys[0] = (mask & 0x20)?_keytest_optimized (RR_UP):0;
		menu_keys[1] = (mask & 0x10)?_keytest_optimized (RR_DOWN):0;
		menu_keys[2] = (mask & 0x08)?_keytest_optimized (RR_LEFT):0;
		menu_keys[3] = (mask & 0x04)?_keytest_optimized (RR_RIGHT):0;
	END_KEYTEST
	menu_keys[4] = (mask & 0x02)?(_keytest (RR_ENTER) | _keytest (RR_ENTER2)):0;
	menu_keys[5] = (mask & 0x01)?_keytest (RR_ESC):0;
}

static inline unsigned char test_key_delay_mask(unsigned short delay, unsigned char mask) {
	unsigned char i, c = 0, r;
	while ((delay != 1) && (!c)) {
		test_key_mask(mask);
		for(i=6, c=0; i--; c += menu_keys[i]);
		if (delay) delay--;
	}
	for(r=0; r<6; r++) if(menu_keys[r]) break;
	while (!(mask & 0x80) && c) {
		test_key_mask(mask);
		for(i=6, c=0; i--; c += menu_keys[i]);
	}
	return r;
}

static inline void title_display(const char * title) {
	unsigned short i, x;
	FontSetSys(F_8x10);
	x = (strlen(title)*4);
	PortSet (pl0, 239, 127);
	for(i=2; i--;) {
		DrawStr (LOGOX - x+i, LOGOY-50, title, A_REVERSE);
		DrawLine (LOGOX - x-3+i, LOGOY-53-i, LOGOX + x+3+i, LOGOY-53-i, A_REVERSE); 
		DrawLine (LOGOX - x-3+i, LOGOY-37-i, LOGOX + x+3+i, LOGOY-37-i, A_REVERSE); 
		DrawLine (LOGOX - x-3+i, LOGOY-53-i, LOGOX - x-3+i, LOGOY-37-i, A_REVERSE); 
		DrawLine (LOGOX + x+3+i, LOGOY-53-i, LOGOX + x+3+i, LOGOY-37-i, A_REVERSE); 
		ScrRectFill (&(SCR_RECT){{LOGOX-x-2, LOGOY-53,LOGOX+x+3, LOGOY-38}}
						 ,&(SCR_RECT){{LOGOX-x-2, LOGOY-53,LOGOX+x+3, LOGOY-38}}, A_XOR);
		PortSet (pl1, 239, 127);
	}
	
}

static inline void small_entry_display(const char * entry, unsigned char nbentries,
									 unsigned char numofentry, char x) {
	unsigned short i, y;
	FontSetSys(F_4x6);
	y = (nbentries*4)-(numofentry*8);
	
	PortSet (pl0, 239, 127);
	for(i=2; i--;) {
		DrawStr (LOGOX - x+i, LOGOY-y, entry, A_XOR);
		PortSet (pl1, 239, 127);
	}
}

static inline void entry_display(const char * entry, unsigned char nbentries,
									 unsigned char numofentry) {
	unsigned short i, x, y;
	FontSetSys(F_6x8);
	x = (strlen(entry)*3);
	y = (nbentries*5)-(numofentry*10);
	
	PortSet (pl0, 239, 127);
	for(i=2; i--;) {
		DrawStr (LOGOX - x+i, LOGOY-y, entry, A_XOR);
		PortSet (pl1, 239, 127);
	}
}

static inline void toggle_entry(unsigned char nbentries, unsigned char numofentry) {
	unsigned short y;
	y = (nbentries*5)-(numofentry*10);
	PortSet (pl0, 239, 127);
		ScrRectFill (&(SCR_RECT){{LOGOX-60, LOGOY-y-1,LOGOX+61, LOGOY-y+7}},
		&(SCR_RECT){{LOGOX-60, LOGOY-y-1,LOGOX+61, LOGOY-y+7}}, A_XOR);
}

static inline char menu(unsigned char nbentries, char choice) {
	unsigned char key=6;

	do {
		if(key == K_DOWN) choice++;
		if(key == K_UP) choice--;
		if(choice==-1) choice = nbentries-1;
		if(choice==nbentries) choice = 0;
		toggle_entry(nbentries,choice);
		key = test_key_delay_mask(0, 0x33);
		toggle_entry(nbentries,choice);
	} while((key != K_ENTER) && (key != K_ESC));

	if(key == K_ESC) return -1;

	return choice;
}

static inline char Menu(const char* Title, const char** Elements, char N, char choice) {
	unsigned char i;
	getplanes();
	clrscrR();
	title_display(Title);
	for (i=N;i--;)
		entry_display(Elements[i], N, i);
	finish_display(1);
	return menu(N,choice);	
}

static inline void small_menu(const char* Title, const char** Elements, const unsigned char * Xpos, char N) {
	unsigned char i;
	getplanes();
	clrscrR();
	title_display(Title);
	for (i=N;i--;)
		small_entry_display(Elements[i], N, i, Xpos[i]);
	finish_display(1);
	test_key_delay_mask(0, 0x03);
}

static inline void trknum2char(unsigned char tracknum) {
	unsigned char unite = 12;
	trackfilename[13] = 0;
	if(tracknum >= 10)
		trackfilename[unite++] = '0' + tracknum/10, tracknum = tracknum%10;
	trackfilename[unite] = '0' + tracknum;
}

static inline void mapPreview(unsigned char small) {
	unsigned short i, j;
	unsigned char z, dimx, dimy;
	memset(mp,0x00,32*4);
	if(small) for(i=22;i--;) memset(mp+i*4,0xFF,2), *(mp+i*4+2) = 0xFC;
	memset(mp+2*32*4,0xFF,32*4);
	dimx = track.xm->dimX;
	dimy = track.xm->dimY;
	for(i = dimx; i--;) {
		for(j = dimy; j--;) {
			z = track.xm->map[i + j*dimx];
			if(small) {
				if (z < nbfx && z >=60) *(unsigned char *)(mp+((i/3)>>3)+((j/3)<<2)) &= ~(0x80 >> ((i/3)&0x7));
				if (z < nbfx && z >=60) *(unsigned char *)(mp+2*32*4+((i/3)>>3)+((j/3)<<2)) &= ~(0x80 >> ((i/3)&0x7));
			} else {
				if (z < 60) *(unsigned char *)(mp+(i>>4)+((j>>1)<<2)) |= (0x80 >> ((i>>1)&0x7));
				if (z < nbfx && z >=60) *(unsigned char *)(mp+2*32*4+(i>>4)+((j>>1)<<2)) &= ~(0x80 >> ((i>>1)&0x7));
			}
		}
	}
	memcpy(mp+32*4, mp, 32*4);
}

static inline void intro(void) {
	unsigned short i;
	
	volatile unsigned short j;

	getplanes();
	m7_DBufToggle();
	
	clrscrG();

	m7sprites_Add1stSprite(m7sprites_LoadSprite(fzeropk0, "sprites\\misc\\orage0"));
	m7sprites_Add1stSprite(m7sprites_LoadSprite(fzeropk0, "sprites\\misc\\orage1"));
	m7sprites_Add1stSprite(m7sprites_LoadSprite(fzeropk0, "sprites\\misc\\orage2"));
	m7sprites_Add1stSprite(m7sprites_LoadSprite(fzeropk0, "sprites\\misc\\orage3"));

	AllSpriteList[0]->dsprite.Xscreen = LOGOX;
	AllSpriteList[0]->dsprite.Yscreen = LOGOY+6;
	AllSpriteList[0]->dsprite.Zscreen = 0;
	m7sprites_Set1stDisplayed(0, 0);

	AllSpriteList[1]->dsprite.Xscreen = LOGOX+8;
	AllSpriteList[1]->dsprite.Yscreen = LOGOY+6;
	AllSpriteList[1]->dsprite.Zscreen = 0;
	m7sprites_Set1stDisplayed(1, 0);

	AllSpriteList[2]->dsprite.Zscreen = 0;
	m7sprites_Set1stDisplayed(2, 0);

	AllSpriteList[3]->dsprite.Xscreen = LOGOX+1;
	AllSpriteList[3]->dsprite.Yscreen = LOGOY+6;
	AllSpriteList[3]->dsprite.Zscreen = 0;
	m7sprites_Set1stDisplayed(3, 0);


		
	m7sprite_DisplayNonClipped(0, pl0, pl1);

	for(j=0;j<400;j++);
	
	m7sprite_DisplayNonClipped(1, pl0, pl1);
	
	for(i = 42; i != 74; i--) {
		for(j=0;j<400;j++);
		AllSpriteList[2]->dsprite.Xscreen = (Cos128Table[i]/6) + LOGOX + 2;
		AllSpriteList[2]->dsprite.Yscreen = (Sin128Table[i]/6) + LOGOY - 14;
		if (i == 0) i = 256;
		m7sprite_DisplayNonClipped(2, pl0, pl1);
	}

	m7sprite_DisplayNonClipped(3, pl0, pl1);

	PortSet (pl0, 239, 127);
	FontSetSys(F_4x6);

	for(i=2; i--;) {
		DrawStr (LOGOX-43, LOGOY+18, "- Orage Studio Presents -", A_NORMAL);
		PortSet (pl1, 239, 127);
	}

	m7sprites_RemoveAllSprites();
	m7sprites_FreeAllSprite();

	test_key_delay_mask(10000, 0x03);
}

static inline char logo_menu(void) {
	unsigned short i;
	unsigned char key=0;
	
	m7sprites_Add1stSprite(m7sprites_LoadSprite(fzeropk0, "sprites\\misc\\fzero"));
	AllSpriteList[0]->dsprite.Xscreen = LOGOX;
	AllSpriteList[0]->dsprite.Yscreen = LOGOY;
	AllSpriteList[0]->dsprite.Zscreen = 0;
	m7sprites_Set1stDisplayed(0, 0);

	getplanes();

	clrscrR();
	m7sprite_DisplayNonClipped(0, pl0, pl1);

	m7sprites_RemoveAllSprites();
	m7sprites_FreeAllSprite();

	PortSet (pl0, 239, 127);

	for(i=2; i--;) {
		FontSetSys(F_4x6);
		DrawStr (LOGOX-52, LOGOY-55, "Original Game by NINTENDO ", A_REVERSE);
		DrawStr (CALCSTRX, LOGOY+3, CALCSTR, A_REVERSE);
		DrawStr (LOGOX-52, LOGOY+25, "Powered by Mode7Engine v1.1", A_REVERSE);
		DrawStr (LOGOX-46, LOGOY+32, " 2003-2005 Lionel Anton", A_REVERSE);
		FontSetSys(F_6x8);
		DrawStr (LOGOX-35, LOGOY+13, "PRESS ENTER", A_REVERSE);
		PortSet (pl1, 239, 127);
	}

	m7_DBufToggle();

	do {
		key = test_key_delay_mask(4000, 0x03);
		PortSet (pl0, 239, 127);
		for(i=2; i--;) {
			ScrRectFill (&(SCR_RECT){{LOGOX-35, LOGOY+12,LOGOX+31, LOGOY+20}}
						 ,&(SCR_RECT){{LOGOX-35, LOGOY+12,LOGOX+31, LOGOY+20}}, A_XOR); 
			PortSet (pl1, 239, 127);
		}
	} while(key == K_NO_PRESS);

	if(key == K_ESC) return -1;
	return 0;
}

static inline char main_menu(void) {
	return Menu(" MAIN MENU ",(const char *[]){"Grand Prix","Versus","Time Attack","Options","Quit"},5,0);
}

static inline char vs_menu(void) {	
	return Menu(" VS BATTLE ",(const char *[]){"Start as Master","Connect to Master"},2,0);
}

static inline void connect_menu(void) {	
	getplanes();
	clrscrR();
	title_display(" CONNECTING ");
	entry_display("Waiting for other", 3, 0);
	entry_display("player to connect", 3, 1);
#ifdef USB_LINK_IMPLEMENTED
	if(linktype)
		entry_display("Press [ON] to abort", 3, 2);
	else
#endif
		entry_display("Press [ESC] to abort", 3, 2);
	finish_display(1);
}

static const char * gp_str[] = {"Ruby","Sapphire","Emerald","Diamond","Black"};
static 	const char * completed_str[] = {"","BRONZE","SILVER","GOLD"};

static inline char gp_menu(char completed) {
	unsigned char i,j;
	getplanes();
	clrscrR();
	title_display(" GRAND PRIX ");
	for(i=save->grandprix;i--;) {
		entry_display(gp_str[i], save->grandprix, i);
		if(completed) {
			FontSetSys(F_4x6);
			PortSet (pl0, 239, 127);
			for(j=2; j--;) {
				DrawStr (LOGOX - 60+j, LOGOY-((save->grandprix*5)-(i*10))+1, completed_str[save->gpcompleted[mparam.cpulvl+i*5]], A_XOR);
				PortSet (pl1, 239, 127);
			}
		}
	}
	finish_display(1);
	return menu(save->grandprix,mparam.gpnum);	
}

const char * class_str[] = {"Novice","Standard","Expert","Master","Ultimate"};

static inline char class_menu(void) {
	return Menu(" CLASS ",class_str,save->class,0);
}


static inline long checksum(unsigned char * data, unsigned short size) {
	long ret = 0;
	for(;size--;) ret += *(data++);
	return ret;
}

static const char * machines_str[] = {"Blue Falcon", "Golden Fox", "Wild Goose", "Fire Stingray"};

static inline void ghostunload_menu(void){
	ghostloaded = 0;
	mparam.nbcpu = 0;
	Free(gparam.data);
}

static inline void time2str(unsigned short timeval, char * timestr) {
	unsigned short t = (timeval)%(256/GAME_SPEED);
	timestr[7] = '0' + ((t*100)/32)%10;
	timestr[6] = '0' + ((t*100)/32)/10;
	t = (timeval)/(256/GAME_SPEED)%60;
	timestr[4] = '0' + t%10;
	timestr[3] = '0' + t/10;
	t = (timeval)/(256/GAME_SPEED)/60;
	timestr[1] = '0' + t%10;
	timestr[0] = '0' + t/10;
}

static inline void ghostload_menu(char staffg) {
	ARCH *ghost;
	
	char ghostfilename[] = "stffghst\\fghstm00";
	unsigned char tnum = (5 * mparam.gpnum) + mparam.tracknum;
	ghostfilename[15] += tnum/10;
	ghostfilename[16] += tnum%10;
	if(staffg) {
		if(!(ghost = aopen (fzerodt0, ghostfilename))) goto error;
	} else {
		if(!(ghost = afopen (SYMSTR(ghostfilename+9)))) goto error;
	}
	unsigned char * ghostptr = ghost->apos;

	if(memcmp(ghostptr, GHOSTFILE_SIGNATURE, 8)) goto not_compat;
	ghostptr += 8;
	if (*((char *)ghostptr) != tnum) goto not_compat;
	ghostptr += 1;
	unsigned char * name = ghostptr;
	ghostptr += 7;
	gparam.time = *((short *)ghostptr);
	ghostptr += 2;
	gparam.selected_machine = *(ghostptr++);
	gparam.balance = *(ghostptr++);
	unsigned short size;
	gparam.data = uncompress(ghostptr, &size);

	aclose(ghost);
	ghostloaded = 1;
	mparam.nbcpu = 1;
	mparam.cpulvl = -1;

	getplanes();
	clrscrR();
	title_display(" GHOST ");
	entry_display(name, 3, 0);
	char time[] = "00'00\"00";
	time2str(gparam.time, time);
	entry_display(time, 3, 1);
	entry_display(machines_str[gparam.selected_machine], 3, 2);
	finish_display(1);
	if (test_key_delay_mask(0,0x3) == K_ESC) ghostunload_menu();
	return;
	not_compat:
	Menu(" ERROR ",(const char *[]){"No ghost compatible"},1,0);
	ghostloaded = 0;
	mparam.nbcpu = 0;
	aclose(ghost);
	return;
	error:
	Menu(" ERROR ",(const char *[]){"No ghost found"},1,0);
}

char archive_file(SYM_STR SymName, unsigned long Size) {
  if(!EM_findEmptySlot(Size)) {
  	EM_GC(FALSE);
		aclose(fzeropk0);
		aclose(fzeropk1);
		aclose(fzeropk2);
		aclose(fzerodt0);
		fzeropk0 = afopen (SYMSTR("fzeropk0"));
		fzeropk1 = afopen (SYMSTR("fzeropk1"));
		fzeropk2 = afopen (SYMSTR("fzeropk2"));
		fzerodt0 = afopen (SYMSTR("fzerodt0"));
  }
  if(EM_moveSymToExtMem (SymName, HS_NULL)) return 0;
  return -1;
}

static inline char preview_menu(unsigned char ghostload) {
	if(ghostload) if(ghostloaded) ghostunload_menu();
	m7sprites_Add1stSprite(m7sprites_LoadSprite(fzeropk0, fstpath[0])); // add dummy sprite
	AllSpriteList[0]->dsprite.Zscreen = 0;
	AllSpriteList[0]->dsprite.width = 4;
	AllSpriteList[0]->dsprite.height = 32;
	AllSpriteList[0]->dsprite.img = mp;
	AllSpriteList[0]->dsprite.mask = 	mp + 32*2*4;
	AllSpriteList[0]->dsprite.Xscreen = LOGOX+18-((track.xm->dimX)>>2);
	AllSpriteList[0]->dsprite.Yscreen = LOGOY+24-((track.xm->dimY)>>2);
	mapPreview(0);
	preview_restart:
	getplanes();
	clrscrR();
	title_display(track.trackName);
	finish_display(1);

	unsigned char tnum = (5 * mparam.gpnum) + mparam.tracknum;

	if(ghostload) {
		if (!ghostloaded) {
			small_entry_display("LOAD GHOST : [\x12]", 0,4, 30);
			if(save->staffghostunlocked[tnum-1]) {
				small_entry_display("STAFF",8,0,-20);
				small_entry_display("GHOST",8,1,-20);
				small_entry_display("[\x13]",8,2,-25);
			}
		}
		else
			small_entry_display("UNLOAD GHOST : [\x11]", 0,4, 34);
	}
	char s[4] = "0/5";
	*s += mparam.tracknum;
	entry_display((const char*)s, 0, 2);
	toggle_entry(0, 2);
	m7sprite_Display1stPlane(0, pl0, pl1);

	char key = test_key_delay_mask(0, ghostload?0x2F:0x03);

	if (!ghostloaded) {
		if (key == K_RIGHT) {
			ghostload_menu(0);
		} else if (save->staffghostunlocked[tnum-1] && key == K_UP) {
			ghostload_menu(1);
		}
	} else {
		if (key == K_LEFT || key == K_ESC) {
			ghostunload_menu();
		}			
	}

	if(key == K_RIGHT || key == K_LEFT || key == K_UP)
		goto preview_restart;

	m7sprites_RemoveAllSprites();
	m7sprites_FreeAllSprite();
	

	return key;
}

static inline char trk_menu(char select) {
	ARCH * trackFile;
	char i;
	char * s[5];

	if (select < 1) {
		for(i=1;i<=5;i++) {
			trknum2char((5 * mparam.gpnum) + i);
			trackFile = aopen (fzerodt0, trackfilename);
			s[i-1] = trackFile->apos;
			aclose (trackFile);
		}
		trkchoice:
		i = 1 + Menu(" TRACK ",(const char **)s,5,mparam.tracknum-1);
		if(!i) return -1;
		mparam.tracknum = i;
		if (select == -1) return 0;
	} else {
		mparam.tracknum = select;
	}
	trknum2char((5 * mparam.gpnum) + mparam.tracknum);
	
	getplanes();
	clrscrR();
	title_display(" TRACK ");
	entry_display("loading...", 1, 0);
	finish_display(1);
	LoadTrack(fzerodt0);

	if (preview_menu(select == -2) == K_ESC) { FreeTrack(); if(select>0) return -1; goto trkchoice; }

	mapPreview(1);

	return 0;
}


char balance_menu(void) {
	unsigned char key;
	unsigned char *pl00, *pl01, *pl10, *pl11;
	
	pl00 = (unsigned char *)Malloc(LCD_SIZE);
	pl01 = (unsigned char *)Malloc(LCD_SIZE);
	getplanes();
	clrscrR();
	pl10 = pl0;
	pl11 = pl1;
	m7_DBufToggle();
	getplanes();
	clrscrR();
	title_display(machines_str[player->selected_machine]);
	finish_display(0);
	char s[20];
	sprintf(s,"      WEIGHT %u",stats[player->selected_machine][0]);
	entry_display(s, 5, 0);
	sprintf(s,"        BODY %c   ",stats[player->selected_machine][1]);
	entry_display(s, 5, 1);
	sprintf(s,"       BOOST %c   ",stats[player->selected_machine][2]);
	entry_display(s, 5, 2);
	sprintf(s,"        GRIP %c   ",stats[player->selected_machine][3]);
	entry_display(s, 5, 3);
	ScrRectFill (&(SCR_RECT){{LOGOX-60, LOGOY+19,LOGOX+61, LOGOY+27}},
		&(SCR_RECT){{LOGOX-60, LOGOY+19,LOGOX+61, LOGOY+27}}, A_XOR);
	toggle_entry(4,4);
	for(key = 40; key--;) DrawLine(LOGOX-58+(key*3), LOGOY+21, LOGOX-58+(key*3), LOGOY+25, A_NORMAL);
	small_entry_display("ACCELERATION                          SPEED", 0,4, 60);
	PortSet (pl0, 239, 127);
	memcpy(pl00, pl0, LCD_SIZE);
	memcpy(pl01, pl1, LCD_SIZE);
	
	m7sprites_AddSprite(m7sprites_LoadSprite(fzeropk0, scaledpath[player->selected_machine]));
	AllSpriteList[0]->dsprite.Xscreen = LOGOX - 35;
	AllSpriteList[0]->dsprite.Yscreen = LOGOY + 3;
	AllSpriteList[0]->dsprite.Zscreen = 0;

	do {
		m7sprites_Set1stDisplayed(0, AllSpriteList[0]->dir);

		memcpy(pl0, pl00, LCD_SIZE);
		memcpy(pl1, pl01, LCD_SIZE);
		
		ScrRectFill (&(SCR_RECT){{LOGOX-60+player->balance, LOGOY+19,LOGOX-59+player->balance, LOGOY+27}},
		&(SCR_RECT){{LOGOX-60+player->balance, LOGOY+19,LOGOX-59+player->balance, LOGOY+27}}, A_XOR);
		
		m7sprite_DisplayNonClipped(0, pl0, pl1);
		memcpy(pl10, pl0, LCD_SIZE);
		memcpy(pl11, pl1, LCD_SIZE);
	
		AllSpriteList[0]->dir-=8;
		AllSpriteList[0]->dir%=NBDIR;
	
		test_key_delay_mask(300, 0x00);
		key = test_key_delay_mask(2, 0x8F);
		if (key == K_LEFT && player->balance) player->balance-=3;
		if (key == K_RIGHT && (player->balance != 120)) player->balance+=3;

	}while(key != K_ENTER && key != K_ESC);
	m7sprites_RemoveAllSprites();
	m7sprites_FreeAllSprite();
	Free(pl00);
	Free(pl01);
	
	return (test_key_delay_mask(0, 0x03) == K_ESC)?-1:0;
}

static inline char machine_menu(void) {
	char ret;

	if((ret = Menu(" MACHINE ", machines_str, 4, player->selected_machine)) != -1)
		player->selected_machine = ret;
	return ret;
}

static inline char timeattack_menu(void) {
	return Menu(" TIME ATTACK ",(const char *[]){"Time Attack","Records"},2,0);	
}

static inline void keys_menu(void) {
	small_menu(" KEYS ", (const char *[]){"On/Off : [ON]", "Turn : [\x11] or [\x12]", "Shorten/Lengthen Jump : [\x13]/[\x14]",
						"Attack : [\x11][\x11] or [\x12][\x12]",	"Go : "GO_STR,	BREAK_STR, BOOST_STR, "Menu : [ESC]"},
						(const unsigned char []){55,55,55,55,55,55,55,55},
						8);
}

static inline void contrast_menu(void) {
	char choice = 0;
	do {
		getplanes();
		clrscrR();
		choice = Menu(" CONTRAST ",(const char *[]){"Up","Down"},2,choice);
		if(!choice)
			OSContrastUp ();
		else if(choice == 1)
			OSContrastDn ();
 }while(choice != -1);
}

static inline void credits_menu(void) {
	const unsigned char * staffroll[] = {
		"",
		"",
		"",
		"",
		"",
		"F-ZERO for TI68k",
		"Version 1.00",
		"Powered by",
		"Mode7Engine v1.1",
		" 2003-2005 Lionel Anton",
		"",
		"",
		"",
		"- Programmer -",
		"",
		"Lionel Anton (LionelA)",
		"anton_lionel@yahoo.fr",
		"",
		"",
		"",
		"- Graphics design -",
		"",
		"Machine sprites",
		"map themes",
		"and ground tiles",
		"by Nintendo ",
		"",
		"Road, Border tiles",
		"background skies and",
		"other graphics by",
		"Lionel Anton",
		"",
		"",
		"",
		"- Track design -",
		"",
		"LionelA",
		"",
		"Special thanks to",
		"naPO who sent me a lot of",
		"drawings for the tracks",
		"",
		"thanks to all people who",
		"sent me track drawings",
		"that I couldn't include",
		"in the game",
		"",
		"",
		"",
		"- PC map editor -",
		"",
		"Pen^2",
		"",
		"",
		"",
		"- Web ghost service -",
		"",
		"FpgForce",
		"http://fpgforce.dyndns.org",
		"/~lionela/",
		"",
		"",
		"",
		"- Staff ghosts -",
		"",
		"LionelA",
		"",
		"",
		"",
		"- Beta tester -",
		"",
		"JfG",
		"",
		"",
		"",
		"- Special Thanks -",
		"(alphabetical order)",
		"",
		"ExtendeD",
		"Fisch2",
		"FpgForce",
		"geogeo",
		"JfG",
		"Jonathan Holbrook",
		"Kevin Kofler",
		"Lionel Debroux",
		"naPO",
		"Pen^2",
		"Pollux",
		"Ximoon",
		"",
		"",
		"An Orage Studio software",
		"F-ZERO for TI68k",
		" 2003-2005 Lionel Anton",
		"",
		"",
		"",
		"",
		""
	};
	unsigned char ** lineptr = (unsigned char **)staffroll, i, key;
	unsigned short k;
	for(k=(sizeof(staffroll)/4)-5;k--;) {
		getplanes();
		clrscrR();
		title_display(" CREDITS ");
		for(i=0;i<5;i++) {
			entry_display(*(lineptr+i), 5, i);
		}
		finish_display(1);
		lineptr++;
		key = test_key_delay_mask(3000, 0x93);
		if(key == K_ENTER || key == K_ESC) break;
	}
	test_key_delay_mask(2,0x03); // wait for the user to release ESC or ENTER
}

static inline void link_menu(void) {
	char lt,ch = 0;

	for(;;) {
		ch = Menu(" LINK PLAY ",(const char *[]){"CPU Players", "CPU level", "Link Type"},3, ch);
		switch(ch) {
			case 0: mparam.nbcpu = Menu(" CPU PLAYERS ",(const char *[]){"0","1","2"},3,save->cpuplayers);
				if (mparam.nbcpu == -1) mparam.nbcpu = save->cpuplayers;
				save->cpuplayers = mparam.nbcpu;
				break;
			case 1: mparam.cpulvl = Menu(" CPU LEVEL ",(const char *[]){"Novice","Standard","Expert","Master","Ultimate"},save->class,save->cpulevel);
				if (mparam.cpulvl == -1) mparam.cpulvl = save->cpulevel;
				save->cpulevel = mparam.cpulvl;
				break;
			case 2:
				lt = Menu(" LINK TYPE ",(const char *[]){"Normal","USB"},
				#ifdef USB_LINK_IMPLEMENTED
					TI89_TITANIUM?2:1
				#else
					1
				#endif	
				,linktype);
				if (lt == -1) lt = linktype;
				linktype = lt;
				break;		
			default:
				goto finlink;
		}
	}
	finlink:
}




static inline void save_menu(void) {
	char ret;
	ret = Menu(" AUTO ARCHIVE ",(const char *[]){"On", "Off"},2, save->autoarchive);
	if (ret != -1) save->autoarchive = ret;
}

static unsigned char inputname[6] = {0,0,0,0,0,0};
static inline void input_name_menu(unsigned char * name, unsigned char newrecord) {
	unsigned char key=6;
	unsigned char x,i;
	unsigned char alpha[1+26+26+10];
	volatile unsigned short delay, d;
	alpha[0] = ' ';
	for(x=0;x<26;x++) { alpha[1+x] = 'A' + x; alpha[1+26+x] = 'a' + x; }
	for(x=0;x<10;x++) { alpha[1+26+26+x] = '0' + x; }
	x = 0;

	do {
		getplanes();
		clrscrR();
		if(newrecord == 1)
			title_display(" NEW RECORD ");
		else if(newrecord == 2)
			title_display(" NAME ");
		else
			title_display(" GHOST SAVE ");
		entry_display("Enter your name", 6, 1);
		switch(key) {
			case K_LEFT: if(x>0) x--; break;
			case K_RIGHT: if(x<5) x++; break;
			case K_DOWN: if(!inputname[x]--) inputname[x] = 26+26+10; break;
			case K_UP: inputname[x]++; inputname[x] %= 1+26+26+10; break;
		}
		for(i=6;i--;) { name[i]=alpha[inputname[i]]; }
		entry_display(name, 6, 3);
		small_entry_display("Use Up, Down, Left, Right and Enter", 8, 7, 60);
		finish_display(1);
		
		PortSet (pl0, 239, 127);
		ScrRectFill (&(SCR_RECT){{LOGOX-17+x*6, LOGOY-((6*5)-(3*10))-1,LOGOX-17+x*6+6, LOGOY-((6*5)-(3*10))+7}},
		&(SCR_RECT){{LOGOX-17+x*6, LOGOY-((6*5)-(3*10))-1,LOGOX-17+x*6+6, LOGOY-((6*5)-(3*10))+7}}, A_XOR);
		if(key==6) delay = 2000;
		else for(d=delay;d--;);
		key = test_key_delay_mask(0 , 0xBE);
		if(delay>100) delay--;
		
	} while(key != K_ENTER);
	test_key_delay_mask(2,0x02); // wait for the user to release Enter

	for(i=6;i--;) { if(name[i] != ' ' || !i) break; else name[i] = '\0'; }
}

static inline void options_menu(unsigned char all) {
	char choice = 0, nb;
	unsigned char i;
	unsigned char name[7] = {0,0,0,0,0,0,0};
	nb = all?7:3;

	do {
		choice = Menu(" OPTIONS ",(const char *[]){"Show Keys","Contrast","Credits",
									"Change Name","Link Play","Auto Archive Save","Restore Default"},nb,choice);			
		
		switch(choice) {
			case 0:
				keys_menu();
				break;
			case 1:
				contrast_menu();
				break;
			case 2:
				credits_menu();
				break;
			case 3:
				for(i=6;i--;) inputname[i] = save->inputname[i];
				input_name_menu(name, 2);
				for(i=6;i--;) save->inputname[i] = inputname[i];
				break;
			case 4:
				link_menu();
				break;
			case 5:
				save_menu();
				break;
			case 6:
				save->autoarchive = 0;
				save->linktype = 0;
				save->cpuplayers = 0;
				save->cpulevel = 0;
				break;
		}
	}while(choice != -1);
}

static inline char pause_menu(void) {
	char  ret;
	test_key_delay_mask(2,0x01); // wait for the user to release ESC
	pause_restart:
	ret = Menu(" PAUSE ",(const char *[]){"Resume","Retry","Options","End"},4,0);
	if (ret==2) {
		options_menu(0);
		goto pause_restart;
	}
	return ret;
}

static inline char abort_menu(void) {
	return Menu(" ABORT ? ",(const char *[]){"No","Yes"},2,0);
}

static inline char retry_menu(void) {
	return Menu(" RETRY ? ",(const char *[]){"No","Yes"},2,0);
}

static inline void rank_menu(void) {
	const char * ranks[] = {"1ST PLACE !","2ND PLACE","3RD PLACE","4TH PLACE..."};
	getplanes();
	clrscrR();
	title_display(" RESULTS ");
	entry_display(machines_str[player->selected_machine], 6, 0);
	entry_display(ranks[rank[1-linkmode]], 4, 1);
	char totaltime[] = "Total Time: 00'00\"00";
	char bestlap[] = "Best Lap  : 00'00\"00";
	if(!linkenabled || (linkenabled && rank[1-linkmode] == 0)) {
		time2str(machines[1-linkmode].totaltime, totaltime+12);
		time2str(machines[1-linkmode].bestlap, bestlap+12);
		entry_display(totaltime, 4, 3);
		entry_display(bestlap, 4, 4);
	}
	finish_display(1);
	test_key_delay_mask(0, 0x02);
}

static unsigned short ranked_points[4];

static inline void race_rank_menu(void) {
	unsigned char i, j=0, d=0,s[] = "0";
	getplanes();
	clrscrR();
	title_display(" RESULTS ");
	entry_display(track.trackName, 5, 0);
	for(i=nb_machines;i--;) {
		if(rank[i] == j) {
			small_entry_display(machines_str[sel_machines[i]], nb_machines-2, d, 60);
			if(j == 4) small_entry_display("DISQUALIFIED", nb_machines-2, d, -16), rank[i] = d, ranked_points[sel_machines[i]] = 0;
			else *s = '1'+j,small_entry_display(s, nb_machines-2, d, -55), j++, ranked_points[sel_machines[i]] = (4-d)*15 + 40;
			i=nb_machines;
			d++;
		} else if(!i && d < nb_machines) {
			i=nb_machines;
			j++;
		}
	}
	finish_display(1);
	test_key_delay_mask(0, 0x02);
}

void fillsend(void) {
	kbd_Scan();
	// send : [UP DOWN LEFT RIGHT GO BREAK BOOST MENU]
	send = 0;
	if(UP)send|=0x80;
	if(DOWN)send|=0x40;
	if(LEFT)send|=0x20;
	if(RIGHT)send|=0x10;
	if(GO)send|=0x08;
	if(BREAK)send|=0x04;
	if(BOOST)send|=0x02;
	if(MENU)send|=0x01;
}

static inline void main_quit(void);

static long start_positions[MAX_MACHINES*2];
static inline void race_init (unsigned char begin) {
	clearAllR();

		setWayPoints();

	lapbuff[6] = '0' + track.nbLaps;
/*
	FILE * f;
	f=fopen("wpm","wb");
	unsigned short zozo,zaza,toto;
	for(zozo = 0;zozo<track.xm->dimY;zozo++) {
		for(zaza = 0;zaza<track.xm->dimX;zaza++) {
			toto = WPM[zaza+(zozo*track.xm->dimX)];
			if(toto < 0x8000) // 0xFFFE
				fprintf(f,"%02x ",toto);
			else if (toto == 0xFFFE)
				fprintf(f,"## ");
			else if (toto < 0xFFFE)
				fprintf(f,"%02x'",toto&0x00FF);
			else
				fprintf(f,"   ");
		}
		fprintf(f,"\n");
	}

	fclose(f);
Exit(0);	
*/
	LoadSky();

	unsigned char timeattackflag = 0;
	long startoffsetx, startoffsety;
	startoffsetx = (1 << (3+SINCOS+ZOOM)) + (track.startx << (4+SINCOS+ZOOM));
	if(track.direction == 1 || track.direction == 3) startoffsetx -= (1 << (2+SINCOS+ZOOM));

	startoffsety = (1 << (3+SINCOS+ZOOM)) + (track.starty << (4+SINCOS+ZOOM));
	if(track.direction == 0 || track.direction == 2) startoffsety += (1 << (2+SINCOS+ZOOM));

	if(linkenabled) { // link play
		nb_machines = 2 + mparam.nbcpu;
	} else { // GP , Time Attack
		nb_machines = 1 + mparam.nbcpu;
		if(mparam.nbcpu <= 1) // time attack
			timeattackflag = 1;
	}

	switch(nb_machines) {
		case 4:
			start_positions[6] = (1 << (3+SINCOS+ZOOM));
			start_positions[7] = (1 << (5+SINCOS+ZOOM));
		case 3:
			start_positions[4] = -(1 << (3+SINCOS+ZOOM));
			start_positions[5] = (1 << (4+SINCOS+ZOOM)) + (1 << (3+SINCOS+ZOOM));
		case 2:
			if(!timeattackflag) {
				start_positions[0] = -(1 << (3+SINCOS+ZOOM));
				start_positions[1] = (1 << (3+SINCOS+ZOOM));
				start_positions[2] = (1 << (3+SINCOS+ZOOM));
				start_positions[3] = (1 << (4+SINCOS+ZOOM));
				break;
			}
		case 1:
			start_positions[0] = 0;
			start_positions[1] = (1 << (3+SINCOS+ZOOM));
			if(timeattackflag) {
				start_positions[2] = start_positions[0];
				start_positions[3] = start_positions[1];
			}
		default:;
	}

	unsigned char i;
	short swaptemp;
	for(i=nb_machines; i--;) {
		if (track.direction == 1 || track.direction == 2) {
			start_positions[i*2] = -start_positions[i*2];
			start_positions[i*2+1] = -start_positions[i*2+1];
		}
		if (track.direction == 1 || track.direction == 3) {
				swaptemp = start_positions[i*2];
				start_positions[i*2] = start_positions[i*2+1];
				start_positions[i*2+1] = -swaptemp;	
		}
		start_positions[i*2] += startoffsetx;
		start_positions[i*2+1] += startoffsety;
		if(begin == 1) rank[i] = nb_machines - i - 1;
	}
	unsigned char j=4, k;
	i = 1;

	sel_machines[0] = mparam.selected_machine;

	if(linkenabled) {
		sel_machines[1] = sparam.selected_machine;
		i = 2;
	}

	if (timeattackflag)
		sel_machines[1] = gparam.selected_machine, i = 2, rank[1] = 0;


	while(i < (nb_machines)) {
		for(;j--;) {
			for(k=i;k--;)
				if(sel_machines[k] == j) break;
			if(k==0xFF) break;
		}
		sel_machines[i] = j;
		i++;
	}

	track.direction = (((2-(track.direction))%4) * NBDIR)/4;
	setGravity(track.gravity);

	m7_SetMap(track.xm);
	m7_SetSky(track.sky->map0,track.sky->map1);

	xm = track.xm;

}

static inline void race_end(void) {
	FreeTrack();
	FreeSky();
}


static inline unsigned char race(unsigned char begin) {
	unsigned char raceret = 0; // 0: finished, 1: dead/retry, 2: end, 3: link abort

	race_init(begin);
	DrawHUDFrame();

	// convention : AllSpriteList : 0 = shadow, 1 = player machine (1st plane), 2 = energy ball, 3 = fire, 4 = player machine (scaled),
	// convention : AllSpriteList : 5 = map, 6 = powerbar, 7 = other machine, 8 = other machine, 9 = other machine
	
	// shadow
	m7sprites_Add1stSprite(m7sprites_LoadSprite(fzeropk0, "sprites\\shared\\shadow"));
	AllSpriteList[0]->dsprite.Xscreen = SPRITEX;
	AllSpriteList[0]->dsprite.Yscreen = SPRITEY+6;
	AllSpriteList[0]->dsprite.Zscreen = 0;
	m7sprites_Set1stDisplayed(0, 0);
	AddedSprites[0] = 3; // invisible

	// player machine (first plane)
	m7sprites_Add1stSprite(m7sprites_LoadSprite(fzeropk0, fstpath[player->selected_machine]));
	AllSpriteList[1]->dsprite.Xscreen = SPRITEX;
	AllSpriteList[1]->dsprite.Yscreen = SPRITEY;
	AllSpriteList[1]->dsprite.Zscreen = 0;
	m7sprites_Set1stDisplayed(1, 0);
	AddedSprites[1] = 3; // invisible

	// energy ball
	m7sprites_Add1stSprite(m7sprites_LoadSprite(fzeropk0, "sprites\\shared\\energy"));
	AllSpriteList[2]->dsprite.Xscreen = SPRITEX;
	AllSpriteList[2]->dsprite.Yscreen = SPRITEY+6;
	AllSpriteList[2]->dsprite.Zscreen = 0;
	m7sprites_Set1stDisplayed(2, 0);
	AddedSprites[2] = 3; // invisible

	// fire
	m7sprites_Add1stSprite(m7sprites_LoadSprite(fzeropk0, firepath[player->selected_machine]));
	AllSpriteList[3]->dsprite.Xscreen = SPRITEX;
	AllSpriteList[3]->dsprite.Yscreen = SPRITEY+5;
	AllSpriteList[3]->dsprite.Zscreen = 0;
	m7sprites_Set1stDisplayed(3, 0);
	AddedSprites[3] = 3; // invisible

	// player machine (scaled)
	m7sprites_AddSprite(m7sprites_LoadSprite(fzeropk0, scaledpath[player->selected_machine]));
	if(linkenabled && !linkmode) { // slave
		AllSpriteList[4]->x = start_positions[(rank[1]*2)];
		AllSpriteList[4]->y = start_positions[(rank[1]*2)+1];
		sprites_id[1] = 1;
		sprites_id[0] = 7;
		controlled_tab[1] = 0;
		controlled_tab[0] = 1;
		balance_tab[1] = sparam.balance;
		balance_tab[0] = mparam.balance;
	} else {
		AllSpriteList[4]->x = start_positions[(rank[0]*2)];
		AllSpriteList[4]->y = start_positions[(rank[0]*2)+1];
		sprites_id[0] = 1;
		sprites_id[1] = 7;
		controlled_tab[0] = 0;
		controlled_tab[1] = 1;
		balance_tab[1] = sparam.balance;
		balance_tab[0] = mparam.balance;
	}
	AllSpriteList[4]->z = 0;
	AllSpriteList[4]->dir = track.direction;

	// map
	m7sprites_Add1stSprite(1); // dummy
	AllSpriteList[5]->dsprite.Xscreen = HUDXO+40;
	AllSpriteList[5]->dsprite.Yscreen = HUDY+34;
	AllSpriteList[5]->dsprite.Zscreen = 0;
	AllSpriteList[5]->dsprite.width = 4;
	AllSpriteList[5]->dsprite.height = 32;
	AllSpriteList[5]->dsprite.img = mp;
	AllSpriteList[5]->dsprite.mask = 	mp + 32*2*4;
	AddedSprites[5] = 3; // invisible

	// power bar
	m7sprites_Add1stSprite(m7sprites_LoadSprite(fzeropk0, "sprites\\misc\\powerbar"));
	AllSpriteList[6]->dsprite.Xscreen = HUDXO + 121;
	AllSpriteList[6]->dsprite.Yscreen = HUDY+14;
	AllSpriteList[6]->dsprite.Zscreen = 0;
	m7sprites_Set1stDisplayed(6, 0);
	AddedSprites[6] = 3; // invisible

	// other machines (link player)
	if(linkenabled) {
		if(mparam.selected_machine != sparam.selected_machine)
			m7sprites_AddSprite(m7sprites_LoadSprite(fzeropk0, scaledpath[linkmode?sparam.selected_machine:mparam.selected_machine]));
		else
			m7sprites_AddSprite(AllSpriteList[4]->slot);
		AllSpriteList[7]->x = start_positions[linkmode?(rank[1]*2):(rank[0]*2)];
		AllSpriteList[7]->y = start_positions[linkmode?(rank[1]*2)+1:(rank[0]*2)+1];
		AllSpriteList[7]->z = 0;
		AllSpriteList[7]->dir = track.direction;
	}

	// other machines (GP/Versus CPU players or Time Attack ghost)
	unsigned char i, id;
	for(i=mparam.nbcpu;i--;) {
		id = m7sprites_AddSprite(m7sprites_LoadSprite(fzeropk0, scaledpath[sel_machines[nb_machines-i-1]]));
		AllSpriteList[id]->x = start_positions[(rank[nb_machines-i-1]*2)];
		AllSpriteList[id]->y = start_positions[(rank[nb_machines-i-1]*2)+1];
		AllSpriteList[id]->z = 0;
		AllSpriteList[id]->dir = track.direction;
		sprites_id[nb_machines-i-1] = id;
		balance_tab[nb_machines-i-1] = ghostloaded?gparam.balance:mparam.balance;
		controlled_tab[nb_machines-i-1] = mparam.cpulvl+3;
	}

	// Add the machines for the race (gameplay)
	resetMachines();
	addMachines();

	
	// camera init (race intro)
	m7_SetCameraPos2(AllSpriteList[4]->x, AllSpriteList[4]->y);
	m7_SetDir(track.direction);
	m7_Translate(-197);

	unsigned char j;
	id = (strlen(track.trackName)*4);
	j = (strlen(track.trackComment)*2);

	DrawHUD4P();
	
	for(i=0;i<4;i++) {
		PortSet (Planes[i], 239, 127);
		FontSetSys(F_8x10);
		DrawStr (LOGOX - id+(i%2), HUDY+5, track.trackName, A_REVERSE);
		FontSetSys(F_4x6);
		DrawStr (LOGOX - j, HUDY+17, track.trackComment, A_REVERSE);
	}


	// main loop
	m7Link_SetFunc(fillsend);
	m7_InitMoreDisplay(d_animate_cpu);
	m7_SetLoopFunc(m_race_intro);
	m7_LoopInit();
	
	if(linkenabled) {
		if(!m7Link_Connect()) goto raceend;
	}
	
	restart_pause:
	end_race_status = 0;

	m7_LoopStart();
		
	switch(end_race_status) {
		case 1: gameover(); raceret = 1; 	test_key_delay_mask(2,0x01); break; // dead animation
		case 2: finish(); break; // race finished animation
		default:
		if(!linkenabled) {
			pausemenu:
			switch(pause_menu()) {
				case -1:
				case 0: if(boostevent == 2) boostevent = 1; if(revevent == 2) revevent = 1; HudChange = 2; MapChange = 2; LapChange = 2; EnergyChange = 2; RankChange = 2; goto restart_pause;
				case 1: if(retry_menu() != 1) goto pausemenu; raceret = 1; break;
				case 2: options_menu(0); goto pausemenu;
				default: if(abort_menu() != 1) goto pausemenu; raceret = 2;
			}
		} else { test_key_delay_mask(2,0x01); raceret = 3;}
	}


	m7sprites_RemoveAllSprites();
	m7sprites_FreeAllSprite();
	
raceend:
	race_end();


	return raceret;
}

static inline char race_menu(char num, char gpretry) {
//gp : au debut (num == 1) class,gp,machine,balance ; apres (num >1) choix uniquement balance
//versus : master : gp,track,machine,balance ; slave : machine balance
//time attack : gp,track,machine,balance
//records : gp,track
	char gpnum;

	l_class:
		if(num == 1 && !gpretry) // gp
			if((mparam.cpulvl = class_menu()) == -1)
				return 0;
	if(!linkmode && linkenabled) // slave
		goto l_machine;
	l_gp:
		if(num < 2 && !gpretry) {
			if((gpnum = gp_menu(num>0)) == -1) {
				if (num < 1) return 0;
				goto l_class;
			}
			mparam.gpnum = gpnum;
		}
	l_trk:
		if(trk_menu(num) == -1) {
			if (num > 1 || gpretry) { // gp 2nd track or more
				return -1;
			}
			goto l_gp;
		}
	if (num == -1) return 1;
	l_machine:
		if(num < 2 && !gpretry) { // gp first track or time attack or versus
			if(machine_menu() == -1) {
				if(linkenabled && !linkmode) // slave
					return 0;
				FreeTrack();
				goto l_trk;
			}
		}
	if(balance_menu() == -1) {
		if (num > 1 || gpretry) { // gp 2nd track or more
			FreeTrack();
			goto l_trk;
		}
		goto l_machine;
	}
	return 1;
}

static inline void grand_prix(void) {
	unsigned char num, gpretry = 0;
	unsigned char remain[] = "Remaining lives : 0", *s[1];
	s[0] = remain;
	linkmode = 1;
	unsigned char i, j, nm, n[4], pointstr[4], rankplayer=0;
	unsigned short max, points[] = {0,0,0,0};
	char ret, lives = -1;
	ghost_record = 0;
	mparam.nbcpu = 3;
	for(num = 1; num <= 5; num++) {
		gp_ret:
		if((ret = race_menu(num, gpretry)) == -1) { if (abort_menu() == 1)	return; else goto gp_ret; }
		if (ret == 0) return;
		if(lives == -1)
			switch(mparam.cpulvl) {
				case 0: lives = 5;break;
				case 1: lives = 4;break;
				case 2: lives = 3;break;
				case 3: lives = 1;break;
				case 4: lives = 0;break;
			}
		switch(race(num)) { //0: finished, 1: dead/retry, 2: end, 3: link abort (not possible here)
			case 1: gpretry = 1; if(lives--) goto loose_live;
			case 2: goto game_over;
			default:;
		}
		computeRank();
		rank_menu();
		race_rank_menu();
		
		getplanes();
		clrscrR();
		title_display("TOTAL RANKING");
		entry_display(track.trackName, 5, 0);
		for(i=nb_machines;i--;) points[i] += ranked_points[sel_machines[i]], n[i] = 0;
		j = 0;
		sort_rank:
		max = 0;
		nm = 4;
		for(i=nb_machines;i--;) {
			if(!n[i]) if(max <= points[i]) max = points[i], nm = i;
		}
		if(nm != 4) {
			sprintf(pointstr, "%03u", max);
			n[nm] = 1;
			small_entry_display(machines_str[sel_machines[nm]], nb_machines-2, j, 60);
			small_entry_display(pointstr, nb_machines-2, j, -47);
			if(!nm) rankplayer = j;
			j++;
			goto sort_rank;
		}
		
		finish_display(1);
		test_key_delay_mask(0, 0x02);
		// reverse start positions
		for(i=nb_machines;i--;) rank[i] = nb_machines-1-rank[i];
	}
	// gp ending
	if(rankplayer<3) {
		getplanes();
		clrscrR();
		title_display("CONGRATULATIONS");
		entry_display(gp_str[(unsigned char)mparam.gpnum], 5, 0);
		entry_display(class_str[(unsigned char)mparam.cpulvl], 5, 1);
		entry_display("You won the", 5, 2);
		entry_display(completed_str[3-rankplayer], 5, 3);
		entry_display("Cup", 5, 4);
		finish_display(1);
		test_key_delay_mask(0, 0x02);
		if(save->gpcompleted[mparam.cpulvl+mparam.gpnum*5] < 3-rankplayer)
			save->gpcompleted[mparam.cpulvl+mparam.gpnum*5] = 3-rankplayer;
		if(!rankplayer) {	// test for unlockables
			if((save->class-1 == mparam.cpulvl) && save->class < 5) {
				save->class++;
				getplanes();
				clrscrR();
				title_display(" UNLOCKED ");
				entry_display("You have unlocked", 2, 0);
				entry_display("a new class !", 2, 1);
				finish_display(1);
				test_key_delay_mask(0, 0x02);
			}
			if((save->grandprix-1 == mparam.gpnum) && save->grandprix < 5) {
				if(
					 save->grandprix == 3 ||
				   (save->grandprix == 4 &&
				    save->gpcompleted[4] == 3 &&
				    save->gpcompleted[9] == 3 &&
				    save->gpcompleted[14] == 3 &&
				    save->gpcompleted[19] == 3
				   )
				  ) {
					save->grandprix++;
					getplanes();
					clrscrR();
					title_display(" UNLOCKED ");
					entry_display("You have unlocked", 2, 0);
					entry_display("a new grand prix !", 2, 1);
					finish_display(1);
					test_key_delay_mask(0, 0x02);
				}
			}
		}
	}
	
	goto game_over;
	loose_live:
	remain[18] = '0'+lives;
	Menu(" LIVES ",(const char **)s,1,0);
	goto gp_ret;
	game_over:
	Menu(" GAME OVER ",(const char *[]){"Thanks for playing"},1,0);
}

static inline void connection_lost_menu() {
	Menu(" ERROR ",(const char *[]){"Connection closed"},1,0);
}

static inline void versus(void) {
	char ret;
	linkenabled = 1;
	mparam.cpulvl = save->cpulevel;
	mparam.nbcpu = save->cpuplayers;
	ghost_record = 0;
	ret = vs_menu();
	if(ret == -1) return;
	linkmode = 1-ret;
	if(!linkmode) player = &sparam;
	else player = (slavep*)&mparam;
	if(!race_menu(0,0)) return;
	connect_menu();
	if(!m7Link_Connect())	{if(linkmode) FreeTrack(); return;}
	if(linkmode) { // master
		if(!LIO_SendData (&mparam, sizeof(masterp))) {
			if(!LIO_RecvData (&sparam, sizeof(slavep), 100)) {
				goto race;
			}
		}
		FreeTrack();
	}	else {
		if(!LIO_RecvData (&mparam, sizeof(masterp), 20)) {
			if(!LIO_SendData (&sparam, sizeof(slavep))) {
	
				trknum2char((5 * mparam.gpnum) + mparam.tracknum);
				LoadTrack(fzerodt0);
				mapPreview(1);
				
				goto race;
			}
		}
	}
	
	return;

	race:
	if(!m7Link_Connect())	{ FreeTrack(); return;}
	m7Link_Flush();
	switch(race(1)) { //0: finished, 1: dead/retry, 2: end, 3: link abort
		case 3: connection_lost_menu();
		case 1:
		case 2:
		goto vs_end;
		default:;
	}
	computeRank();
	rank_menu();
	vs_end:
	m7Link_Flush();
}

static inline void record_menu(record * recptr, unsigned char bestlap, char hl) {
	unsigned char i;
	char time[] = "00'00\"00";
	getplanes();
	clrscrR();
	title_display(" RECORDS ");
	entry_display(track.trackName, 6, 0);
	entry_display(bestlap?"Best Laps":"Total Times", 6, 1);
	for(i=5;i--;) {
		if(recptr[i].name[0]) {
			small_entry_display(recptr[i].name, 5, i+2, 60);
			small_entry_display(machines_str[recptr[i].machine], 5, i+2, 22);
			time2str(recptr[i].time, time);
			small_entry_display(time, 5, i+2, -28);
		}
	}
	finish_display(1);
	if(hl>=0) {
		unsigned short y;
		y = (5*4)-((hl+2)*8);
		PortSet (pl0, 239, 127);
			ScrRectFill (&(SCR_RECT){{LOGOX-60, LOGOY-y-1,LOGOX+61, LOGOY-y+5}},
			&(SCR_RECT){{LOGOX-60, LOGOY-y-1,LOGOX+61, LOGOY-y+5}}, A_XOR);
	}
}

static inline char ta_restart_menu(void) {
	return Menu(track.trackName,(const char *[]){"Retry","Ghost save","End"},3,0);
}

static inline void ghost_record_failed(void) {
	getplanes();
	clrscrR();
	title_display(" GHOST ERROR ");
	entry_display("Not enough memory", 2, 0);
	entry_display("ghost will not be recorded", 2, 1);
	finish_display(1);
	test_key_delay_mask(0, 0x02);
}

static inline void ghost_record_overburn(void) {
	getplanes();
	clrscrR();
	title_display(" GHOST ERROR ");
	entry_display("Time exceeded during", 2, 0);
	entry_display("recording", 2, 1);
	finish_display(1);
	test_key_delay_mask(0, 0x02);
}

static inline void ghost_save_menu() {
	
	char ghostfilename[] = "fghstm00";
	unsigned char tnum = (5 * mparam.gpnum) + mparam.tracknum;
	ghostfilename[6] += tnum/10;
	ghostfilename[7] += tnum%10;

	SYM_ENTRY *sym;
	unsigned char * compressed, * savefile;
	unsigned short sizeComp;
	compressed = compress((unsigned char*)ghost_record_buffer+2, (ghost_record*2)-2, &sizeComp);
	Free(ghost_record_buffer);
	savefile = (unsigned char*)Malloc(sizeComp+2+20+1+5+1);
	memset(savefile, 0x00, sizeComp+2+20+1+5+1);
	*(short*)savefile = sizeComp+20+1+5+1;
	memcpy((short*)(savefile+2), GHOSTFILE_SIGNATURE, 8);
	*(char*)(savefile+10) = tnum;
	input_name_menu((char*)(savefile+11), 0);
	*(short*)(savefile+18) = machines[0].totaltime;
	*(char*)(savefile+20) = mparam.selected_machine;
	*(char*)(savefile+21) = mparam.balance;
	memcpy(savefile+22,compressed,sizeComp);
	Free(compressed);
	*(unsigned char*)(savefile+sizeComp+2+20) = 0;
	strcpy(savefile+sizeComp+2+20+1,"GHST");
	*(unsigned char*)(savefile+sizeComp+2+20+1+5) = OTH_TAG;
	sym = SymFindPtr(SYMSTR(ghostfilename),0);
	if(sym) {
		if(sym->flags.bits.archived) {
			if(EM_moveSymFromExtMem (SYMSTR(ghostfilename), HS_NULL) == FALSE) Exit(1);
		}
		HeapFree(sym->handle);
	} else {
		if(!(sym=DerefSym(SymAdd(SYMSTR(ghostfilename))))) Exit(1);
	}
	sym->handle = HeapUnlock(Keep(savefile));
	if (archive_file(SYMSTR(ghostfilename), HeapSize(sym->handle))) goto archive_error;
	return;
	archive_error:
	small_menu(
		" ERROR ",
		(const char *[]){
			"Ghost file has not been",
			"archived, there is not",
			"enough free archive mem",
			"",
			"Press ENTER to continue",
		},
		(const unsigned char []){60,60,60,40}, 4
	);
	test_key_delay_mask(0, 0x02);
}

static inline void time_attack(void) {
	mparam.nbcpu = 0;
	record * recptr;
	linkmode = 1;
	unsigned char ghost_save;
	unsigned char * name;
	unsigned char total;
	unsigned short time;
	unsigned short i,j;
	switch(timeattack_menu()) {
		case 0:
			ta_ret:
			ghostloaded = 0;
			if(!race_menu(-2,0)) return;
			//////////
/*			mparam.gpnum = 4;
			mparam.tracknum = 4;
				trknum2char((5 * mparam.gpnum) + mparam.tracknum);
				LoadTrack(fzerodt0);
				//mapPreview(1);
	*/		
			/////////
			
			
			// allouer le buffer pour recorder le ghost
			ghost_save = 0;
			if(!(ghost_record_buffer = (unsigned char *)MallocNoExit(GHOST_BUFFER_SIZE))) ghost_record = 0, ghost_record_failed();
			else ghost_record = 1, ghost_save = 1;
			switch(race(1)) { //0: finished, 1: dead/retry, 2: end, 3: link abort (not possible here)
				case 1: if(ghostloaded) ghostunload_menu(); if(ghost_save) Free(ghost_record_buffer); goto ta_ret;
				case 2: if(ghostloaded) ghostunload_menu(); if(ghost_save) Free(ghost_record_buffer); return;
				default:;
			}
			computeRank();
			rank_menu();
			if(ghostloaded) ghostunload_menu();
			// test new record
			name = NULL;
			time = machines[0].totaltime;
			total = 0;
			new_record_restart:
			recptr = &(save->records[((5 * mparam.gpnum) + mparam.tracknum - 1)*10 + total]);
			for(i=0;i<5;i++) {
				if(recptr[i].name[0]) {
					if(recptr[i].time > time) {
						goto new_record_move;
					}
				} else {
						goto new_record;
				}
			}
			goto next_record;
			new_record_move:
			for(j=4;j>i;j--) {
				memcpy(&recptr[j],&recptr[j-1],sizeof(record));
			}
			new_record:
			recptr[i].machine = player->selected_machine;
			recptr[i].time = time;
			if(!name) {
				name = recptr[i].name;
				input_name_menu(name, 1);
			} else {
				strcpy(recptr[i].name,name);
			}
			record_menu(recptr, total, i);
			test_key_delay_mask(0, 0x02);
			next_record:
			time = machines[0].bestlap;
			if(!total) {total = 5; goto new_record_restart;}
			// test unlock staff ghost
			unsigned char tnum = (5 * mparam.gpnum) + mparam.tracknum;
			if(!save->staffghostunlocked[tnum-1]) {
				ARCH *ghost;
				char ghostfilename[] = "stffghst\\fghstm00";
				ghostfilename[15] += tnum/10;
				ghostfilename[16] += tnum%10;
				if((ghost = aopen (fzerodt0, ghostfilename))) {
					if(machines[0].totaltime <= *(unsigned short*)((ghost->apos)+16) + 256/GAME_SPEED)
						save->staffghostunlocked[tnum-1] = 1;
					aclose(ghost);
				}
			}
			ta_restart_menu_restart:
			switch(ta_restart_menu()) {
				case 0: if(ghost_save) Free(ghost_record_buffer); goto ta_ret;
				case 1: if(!ghost_save) goto ta_restart_menu_restart;
								if(!ghost_record) ghost_record_overburn();
								else {ghost_save = 0; ghost_save_menu(); goto ta_restart_menu_restart;}
				default:;
			}
			if(ghost_save) Free(ghost_record_buffer);
			break;
		case 1:
			rec_restart:
			if(!race_menu(-1,0)) return;
			total = 0;
			rec_display:
			recptr = &(save->records[((5 * mparam.gpnum) + mparam.tracknum - 1)*10 + total]);
			record_menu(recptr, total, -1);
			switch(test_key_delay_mask(0, 0x0D)) {
				case K_RIGHT: total = 5; goto rec_display;
				case K_LEFT: total = 0; goto rec_display;
				default: goto rec_restart;
			}
		default:;
	}
}

static inline void mem_load(void) {
	SYM_ENTRY *sym;
	sym = SymFindPtr(SYMSTR("fzerosav"),0);
	unsigned char * saveptr;
	unsigned short size;
	unsigned char i;
	long chksum;
	if(sym) {
		saveptr = HeapDeref(sym->handle);
		size = *(short*)saveptr;
		saveptr += 2;
		if(*(short*)saveptr != FZERO_VERSION) goto version_error; // may be a save converter in future versions of F-ZERO
		chksum = checksum(saveptr+6, size-2-4-1-5-1);
		saveptr += 2;
		if(*(long*)saveptr != chksum) goto chksum_error;
		saveptr += 4;
		save = (save_struct*)uncompress(saveptr, &size);
		for(i=6;i--;) inputname[i] = save->inputname[i];
		return;
	} else {
		small_menu(
			" WELCOME ",
			(const char *[]){
				"This is the first time you play to",
				"F-ZERO, a save file will be created",
				"",
				"Press ENTER to continue"
			},
			(const unsigned char []){54,60,60,42}, 4
		);
		create_save_file:
		save = (save_struct*)Malloc(sizeof(save_struct));
		unsigned char name[7] = {0,0,0,0,0,0,0};
		input_name_menu(name, 2);
		for(i=6;i--;) save->inputname[i] = inputname[i];
		save->cpuplayers = 0;
		save->cpulevel = 0;
		save->linktype = 0;
		save->autoarchive = 0;
		save->grandprix = 3;
		save->class = 3;
		memset(save->records, 0x00, 250*sizeof(record));
		memset(save->gpcompleted, 0x00, 25);
		memset(save->staffghostunlocked, 0x00, 25);
		return;
	}
	version_error:
	small_menu(
		" ERROR ",
		(const char *[]){
			"Your save is not compatible",
			"with this version of F-ZERO",
			"A new file will be created",
			"The old one will be erased",
			"",
			"Press ENTER to continue",
			"Press ESC to cancel"
		},
		(const unsigned char []){48,48,46,46,60,40,32}, 7
	);
	if(test_key_delay_mask(0, 0x03) == K_ENTER) goto create_save_file;
	else Exit(0);
	chksum_error:
	small_menu(
		" ERROR ",
		(const char *[]){
			"Your save is corrupted",
			"A new file will be created",
			"The old one will be erased",
			"",
			"Press ENTER to continue",
			"Press ESC to cancel"
		},
		(const unsigned char []){38,46,46,60,40,32}, 6
	);
	if(test_key_delay_mask(0, 0x03) == K_ENTER) goto create_save_file;
	Exit(0);
}

static inline void main_quit(void) {
	SYM_ENTRY *sym;
	unsigned char * compressed, * savefile;
	unsigned char autoarchive = save->autoarchive;
	unsigned short sizeComp;
	compressed = compress((unsigned char*)save, sizeof(save_struct), &sizeComp);
	Free(save);
	savefile = (unsigned char*)Malloc(sizeComp+2+2+4+1+5+1);
	*(short*)savefile = sizeComp+2+4+1+5+1;
	*(short*)(savefile+2) = FZERO_VERSION;
	memcpy(savefile+8,compressed,sizeComp);
	Free(compressed);
	*(long*)(savefile+4) = checksum(savefile+8, sizeComp);
	*(unsigned char*)(savefile+sizeComp+2+2+4) = 0;
	strcpy(savefile+sizeComp+2+2+4+1,"FSAV");
	*(unsigned char*)(savefile+sizeComp+2+2+4+1+5) = OTH_TAG;
	sym = SymFindPtr(SYMSTR("fzerosav"),0);
	if(sym) {
		if(sym->flags.bits.archived) {
			if(EM_moveSymFromExtMem (SYMSTR("fzerosav"), HS_NULL) == FALSE) Exit(1);
		}
		HeapFree(sym->handle);
	} else {
		if(!(sym=DerefSym(SymAdd(SYMSTR("fzerosav"))))) Exit(1);
	}
	sym->handle = HeapUnlock(Keep(savefile));
	if(!autoarchive) if (archive_file(SYMSTR("fzerosav"), HeapSize(sym->handle))) goto archive_error;
	return;
	archive_error:
	small_menu(
		" ERROR ",
		(const char *[]){
			"Save file has not been",
			"archived, there is not",
			"enough free archive mem",
			"",
			"Press ENTER to continue",
		},
		(const unsigned char []){60,60,60,40}, 4
	);
	test_key_delay_mask(0, 0x02);
}

char linkbreakfunc(void) {
	return (test_key_delay_mask(2,0x01) == K_ESC);
}

void game(void) {
	char choice;
	char quit = 0;

	m7Link_SetBreakFunc(linkbreakfunc);
	
	intro();
	if(!logo_menu()) {
		mem_load();
		mparam.selected_machine = 0;
		mparam.balance = 60;
		mparam.gpnum = 0;
		mparam.tracknum = 1;
		mparam.nbcpu = 0;
		mparam.cpulvl = 0;
		sparam.selected_machine = 0;
		sparam.balance = 60;
		while (!quit) {
			linkenabled = 0;
			player = (slavep*)&mparam;
			choice = main_menu();
			switch (choice) {
			case 0:
				grand_prix();
				break;
			case 1:
				versus();
				break;
			case 2:
///////////
/*

		save = (save_struct*)Malloc(sizeof(save_struct));
		save->cpuplayers = 0;
		save->cpulevel = 0;
		save->linktype = 0;
		save->autoarchive = 0;
		save->grandprix = 5;
		save->class = 3;
*/
///////////




				time_attack();
				break;
			case 3:
				options_menu(1);
				break;
			case -1:
			case 4:
				main_quit();
				quit = 1;
				break;
			default:;
			}		
		}
	}
}
