// C Header File
// Created 04/02/2004; 09:47:19

#ifndef MODE7_H
#define MODE7_H

#include "Mode7Engine.h"

extern unsigned char linkmode;
extern const signed char * Cos128Table;
extern const signed char * Sin128Table;
extern signed char CosTable[NBDIR];	// contains cos values (useful for faster calculus)
extern signed char SinTable[NBDIR];	// contains sin values
extern short Dir;	// camera direction
extern unsigned char * MapData;
extern short DimX;
extern short DimY;

#if ENABLE_GRAYSCALE == 1
extern unsigned char * Plane0;
extern unsigned char * Plane1;
#else
extern char Virtual[LCD_SIZE];
#endif

// inits
void m7_Init(short SizeHorz, short ViewL, short ViewF);
	// SizeHorz: nb of times that the up is larger than the bottom ?? (try 9)
	// ViewL: angle of view ?? (try 60)
	// ViewF: front distance ?? (try 20)
void m7_ReInit(short SizeHorz, short ViewL, short ViewF);
#if ENABLE_MORE_DISPLAY == 1
void m7_InitMoreDisplay(void (*moreDispFunc)(void));
#endif
#if ENABLE_LARGE_MAPS == 2
void m7_SetMap(ExtraMap * m);	// init of map
#else
void m7_SetMap(Map * m);	// init of map
#endif
#if ENABLE_GRAYSCALE == 1
void m7_GrayInit(void);	// initialization for grayscale (call it after a keyboard init)
void m7_GrayEnd(void);	// end grayscale (call it before a keyboard ending)
void m7_DBufToggle(void); // toggles hidden and active planes
void m7_SetSky(unsigned char *sky0, unsigned char *sky1); // init of sky
#else
void m7_SetSky(unsigned char *sky); // init of sky
#endif

// transformations
void m7_LoadIdentity(void);	// reset position and angle
void m7_Rotate(short angle);	// rotate of angle (specified in number of subdivision previously defined)
void m7_SetDir(unsigned short dir); // set the direction of the camera
void m7_LookAt(short x, short y); // look in the direction of the point (x,y)
void m7_TranslateXY(short x, short y);	// translate x and y position
void m7_TranslateDir(short dist, unsigned short dir);	 // translate dist in the direction dir
void m7_Translate(short dist);	 // translate dist in the current direction
void m7_SetCameraPos(short x, short y); // set the camera at (x << SINCOS,y << SINCOS) real coords
void m7_SetCameraPos2(short x, short y); // set the camera at (x,y)
void m7_GetCameraPos(short * x, short * y); // get the camera position (x,y)

// display loop
void m7_SetLoopFunc(void (*modifFunc)(void));
void m7_LoopInit(void);
void m7_LoopStart(void);
void m7_LoopEnd(void);

#endif
